import platform
from distutils.version import StrictVersion
from Foundation import NSObject
from AppKit import NSFont, NSRegularControlSize, NSSmallControlSize, NSMiniControlSize, \
    NSViewMinXMargin, NSViewMaxXMargin, NSViewMaxYMargin, NSViewMinYMargin, \
    NSViewWidthSizable, NSViewHeightSizable, \
    NSLayoutConstraint, NSLayoutFormatAlignAllLeft, \
    NSLayoutAttributeLeft, NSLayoutAttributeRight, NSLayoutAttributeTop, NSLayoutAttributeBottom, NSLayoutAttributeLeading, NSLayoutAttributeTrailing, \
    NSLayoutAttributeWidth, NSLayoutAttributeHeight, NSLayoutAttributeCenterX, NSLayoutAttributeCenterY, NSLayoutAttributeBaseline, \
    NSLayoutRelationLessThanOrEqual, NSLayoutRelationEqual, NSLayoutRelationGreaterThanOrEqual

try:
    from AppKit import NSLayoutAttributeLastBaseline, NSLayoutAttributeFirstBaseline
except ImportError:
    NSLayoutAttributeLastBaseline = 11
    NSLayoutAttributeFirstBaseline = 12

from vanilla.nsSubclasses import getNSSubclass


class VanillaError(Exception): pass

class VanillaWarning(Warning): pass

# --------------------
# OS Version Constants
# --------------------

macVersion = platform.mac_ver()[0]
if platform.system() != "Darwin":
    macVersion = "0.0"
osVersionCurrent = StrictVersion(macVersion)
osVersion12_0 = StrictVersion("12.0") 
osVersion10_16 = StrictVersion("10.16")  # macOS11 Big Sur seems to be 10.16
osVersion10_15 = StrictVersion("10.15")
osVersion10_14 = StrictVersion("10.14")
osVersion10_13 = StrictVersion("10.13")
osVersion10_12 = StrictVersion("10.12")
osVersion10_11 = StrictVersion("10.11")
osVersion10_10 = StrictVersion("10.10")
osVersion10_9 = StrictVersion("10.9")
osVersion10_8 = StrictVersion("10.8")
osVersion10_7 = StrictVersion("10.7")
osVersion10_6 = StrictVersion("10.6")


# ---------
# Base View
# ---------

class VanillaBaseObject(object):

    frameAdjustments = None

    def __setattr__(self, attr, value):
        _setAttr(VanillaBaseObject, self, attr, value)

    def __delattr__(self, attr):
        _delAttr(VanillaBaseObject, self, attr)

    def _setupView(self, classOrName, posSize, callback=None):
        self._autoLayoutViews = {}
        self._testForDeprecatedAttributes()
        cls = getNSSubclass(classOrName)
        self._nsObject = cls(self)
        self._posSize = posSize
        self._setCallback(callback)
        self._setAutosizingFromPosSize(posSize)

    def _breakCycles(self):
        if hasattr(self, "_target"):
            self._target.callback = None

    def _testForDeprecatedAttributes(self):
        from warnings import warn
        if hasattr(self, "_frameAdjustments"):
            warn(DeprecationWarning("The _frameAdjustments attribute is deprecated. Use the frameAdjustments attribute."))
            self.frameAdjustments = self._frameAdjustments
        if hasattr(self, "_allFrameAdjustments"):
            warn(DeprecationWarning("The _allFrameAdjustments attribute is deprecated. Use the allFrameAdjustments attribute."))
            self.allFrameAdjustments = self._allFrameAdjustments

    def _setCallback(self, callback):
        if callback is not None:
            self._target = VanillaCallbackWrapper(callback)
            self._nsObject.setTarget_(self._target)
            self._nsObject.setAction_("action:")

    def _setAutosizingFromPosSize(self, posSize):
        if posSize == "auto":
            return
        l, t, w, h = posSize
        mask = 0
        if l < 0:
            mask |= NSViewMinXMargin
        if w <= 0 and (w > 0 or l >= 0):
            mask |= NSViewWidthSizable
        if w > 0 and l >= 0:
            mask |= NSViewMaxXMargin
        if t < 0:
            mask |= NSViewMaxYMargin
        if h <= 0 and (h > 0 or t >= 0):
            mask |= NSViewHeightSizable
        if h > 0 and t >= 0:
            mask |= NSViewMinYMargin
        self._nsObject.setAutoresizingMask_(mask)

    def _setFrame(self, parentFrame, animate=False):
        if self._posSize == "auto":
            return
        l, t, w, h = self._posSize
        frame  = _calcFrame(parentFrame, ((l, t), (w, h)))
        frame = self._adjustPosSize(frame)
        if animate:
            self._nsObject.animator().setFrame_(frame)
        else:
            self._nsObject.setFrame_(frame)

    def _adjustPosSize(self, frame):
        if hasattr(self._nsObject, "cell") and self._nsObject.cell() is not None:
            sizeStyle = _reverseSizeStyleMap[self._nsObject.cell().controlSize()]
        else:
            sizeStyle = None
        adjustments = self.frameAdjustments
        if adjustments:
            if sizeStyle is None:
                aL, aB, aW, aH = adjustments
            else:
                aL, aB, aW, aH = adjustments.get(sizeStyle, (0, 0, 0, 0))
            (fL, fB), (fW, fH) = frame
            fL = fL + aL
            fB = fB + aB
            fW = fW + aW
            fH = fH + aH
            frame = ((fL, fB), (fW, fH))
        return frame

    def _getContentView(self):
        return self._nsObject

    def enable(self, onOff):
        """
        Enable or disable the object. *onOff* should be a boolean.
        """
        self._nsObject.setEnabled_(onOff)

    def isVisible(self):
        """
        Return a bool indicating if the object is visible or not.
        """
        return not self._nsObject.isHidden()

    def show(self, onOff):
        """
        Show or hide the object.

        **onOff** A boolean value representing if the object should be shown or not.
        """
        self._nsObject.setHidden_(not onOff)

    def getPosSize(self):
        """
        The position and size of the object as a tuple of form *(left, top, width, height)*.
        """
        return self._posSize

    def setPosSize(self, posSize, animate=False):
        """
        Set the position and size of the object.

        **posSize** A tuple of form *(left, top, width, height)*.

        **animate** A boolean flag telling to animate the transition. Off by default.
        """
        self._posSize = posSize
        if posSize == "auto":
            return
        self._setAutosizingFromPosSize(posSize)
        superview = self._nsObject.superview()
        if superview is not None:
            self._setFrame(superview.frame(), animate)
            superview.setNeedsDisplay_(True)

    def addAutoPosSizeRules(self, rules, metrics=None):
        """
        Add auto layout rules for controls/view in this view.

        **rules** must be a list of rule definitions.
        Rule definitions may take two forms:

        * strings that follow the `Visual Format Language`_
        * dictionaries with the following key/value pairs:

        +---------------------------+-------------------------------------------------------------------+
        | key                       | value                                                             |
        +===========================+===================================================================+
        | *"view1"*                 | The vanilla wrapped view for the left side of the rule.           |
        +---------------------------+-------------------------------------------------------------------+
        | *"attribute1"*            | The attribute of the view for the left side of the rule.          |
        |                           | See below for options.                                            |
        +---------------------------+-------------------------------------------------------------------+
        | *"relation"* (optional)   | The relationship between the left side of the rule                |
        |                           | and the right side of the rule. See below for options.            |
        |                           | The default value is `"=="`.                                      |
        +---------------------------+-------------------------------------------------------------------+
        | *"view2"*                 | The vanilla wrapped view for the right side of the rule.          |
        +---------------------------+-------------------------------------------------------------------+
        | *"attribute2"*            | The attribute of the view for the right side of the rule.         |
        |                           | See below for options.                                            |
        +---------------------------+-------------------------------------------------------------------+
        | *"multiplier"* (optional) | The constant multiplied with the attribute on the right side of   |
        |                           | the rule as part of getting the modified attribute.               |
        |                           | The default value is `1`.                                         |
        +---------------------------+-------------------------------------------------------------------+
        | *"constant"* (optional)   | The constant added to the multiplied attribute value on the right |
        |                           | side of the rule to yield the final modified attribute.           |
        |                           | The default value is `0`.                                         |
        +---------------------------+-------------------------------------------------------------------+

        The `attribute1` and `attribute2` options are:

        +-------------------+--------------------------------+
        | value             | AppKit equivalent              |
        +===================+================================+
        | *"left"*          | NSLayoutAttributeLeft          |
        +-------------------+--------------------------------+
        | *"right"*         | NSLayoutAttributeRight         |
        +-------------------+--------------------------------+
        | *"top"*           | NSLayoutAttributeTop           |
        +-------------------+--------------------------------+
        | *"bottom"*        | NSLayoutAttributeBottom        |
        +-------------------+--------------------------------+
        | *"leading"*       | NSLayoutAttributeLeading       |
        +-------------------+--------------------------------+
        | *"trailing"*      | NSLayoutAttributeTrailing      |
        +-------------------+--------------------------------+
        | *"width"*         | NSLayoutAttributeWidth         |
        +-------------------+--------------------------------+
        | *"height"*        | NSLayoutAttributeHeight        |
        +-------------------+--------------------------------+
        | *"centerX"*       | NSLayoutAttributeCenterX       |
        +-------------------+--------------------------------+
        | *"centerY"*       | NSLayoutAttributeCenterY       |
        +-------------------+--------------------------------+
        | *"baseline"*      | NSLayoutAttributeBaseline      |
        +-------------------+--------------------------------+
        | *"lastBaseline"*  | NSLayoutAttributeLastBaseline  |
        +-------------------+--------------------------------+
        | *"firstBaseline"* | NSLayoutAttributeFirstBaseline |
        +-------------------+--------------------------------+

        Refer to the `NSLayoutAttribute`_ documentation for the information about what each of these do.

        The `relation` options are:

        +--------+------------------------------------+
        | value  | AppKit equivalent                  |
        +========+====================================+
        | *"<="* | NSLayoutRelationLessThanOrEqual    |
        +--------+------------------------------------+
        | *"=="* | NSLayoutRelationEqual              |
        +--------+------------------------------------+
        | *">="* | NSLayoutRelationGreaterThanOrEqual |
        +--------+------------------------------------+

        Refer to the `NSLayoutRelation`_ documentation for the information about what each of these do.

        **metrics** may be either *None* or a dict containing
        key value pairs representing metrics keywords used in the
        rules defined with strings.

        .. _Visual Format Language: https://developer.apple.com/library/archive/documentation/UserExperience/Conceptual/AutolayoutPG/VisualFormatLanguage.html#//apple_ref/doc/uid/TP40010853-CH27-SW1
        .. _NSLayoutAttribute: https://developer.apple.com/documentation/uikit/nslayoutattribute?language=objc
        .. _NSLayoutRelation: https://developer.apple.com/documentation/uikit/nslayoutrelation?language=objc

        """
        _addAutoLayoutRules(self, rules, metrics)

    def move(self, x, y):
        """
        Move the object by *x* units and *y* units.
        """
        posSize = self.getPosSize()
        if posSize == "auto":
            return
        l, t, w, h = posSize
        l = l + x
        t = t + y
        self.setPosSize((l, t, w, h))

    def resize(self, width, height):
        """
        Change the size of the object to *width* and *height*.
        """
        posSize = self.getPosSize()
        if posSize == "auto":
            return
        l, t, w, h = posSize
        self.setPosSize((l, t, width, height))


# ------------
# Base Control
# ------------

_sizeStyleMap = {
    "regular": NSRegularControlSize,
    "small": NSSmallControlSize,
    "mini": NSMiniControlSize
}

_reverseSizeStyleMap = {
    NSRegularControlSize: "regular",
    NSSmallControlSize: "small",
    NSMiniControlSize: "mini"
}

class VanillaBaseControl(VanillaBaseObject):

    def _setSizeStyle(self, value):
        value = _sizeStyleMap[value]
        self._nsObject.cell().setControlSize_(value)
        font = NSFont.systemFontOfSize_(NSFont.systemFontSizeForControlSize_(value))
        self._nsObject.setFont_(font)

    def setTitle(self, title):
        """
        Set the control title.

        **title** A string representing the title.
        """
        self._nsObject.setTitle_(title)

    def getTitle(self):
        """
        Get the control title.
        """
        return self._nsObject.title()

    def isEnabled(self):
        """
        Return a bool indicating if the object is enable or not.
        """
        return self._nsObject.isEnabled()

    def set(self, value):
        raise NotImplementedError

    def get(self):
        raise NotImplementedError

    def bind(self, key, callback):
        raise NotImplementedError


# -------------------
# Sub-View Management
# -------------------

def _recursiveSetFrame(view):
    for subview in view.subviews():
        if hasattr(subview, "vanillaWrapper"):
            obj = subview.vanillaWrapper()
            if obj is not None and hasattr(obj, "_posSize"):
                obj.setPosSize(obj.getPosSize())
        _recursiveSetFrame(subview)

def _setAttr(cls, obj, attr, value):
    if hasattr(value, "getPosSize") and value.getPosSize() == "auto":
        view = value._nsObject
        view.setTranslatesAutoresizingMaskIntoConstraints_(False)
        obj._autoLayoutViews[attr] = view
    if isinstance(value, VanillaBaseObject) and hasattr(value, "_posSize"):
        assert not hasattr(obj, attr), "can't replace vanilla attribute"
        view = obj._getContentView()
        frame = view.frame()
        value._setFrame(frame)
        view.addSubview_(value._nsObject)
        _recursiveSetFrame(value._nsObject)
    #elif isinstance(value, NSView) and not attr.startswith("_"):
    #    assert not hasattr(obj, attr), "can't replace vanilla attribute"
    #    view = obj._getContentView()
    #    view.addSubview_(value)
    super(cls, obj).__setattr__(attr, value)

def _delAttr(cls, obj, attr):
    if hasattr(obj, "_autoLayoutViews"):
        if attr in obj._autoLayoutViews:
            del obj._autoLayoutViews[attr]
    value = getattr(obj, attr)
    if isinstance(value, VanillaBaseObject):
        value._nsObject.removeFromSuperview()
    #elif isinstance(value, NSView):
    #    value.removeFromSuperview()
    super(cls, obj).__delattr__(attr)


# -------------------
# Auto Layout Support
# -------------------

_layoutAttributeMap = dict(
   left=NSLayoutAttributeLeft,
   right=NSLayoutAttributeRight,
   top=NSLayoutAttributeTop,
   bottom=NSLayoutAttributeBottom,
   leading=NSLayoutAttributeLeading,
   trailing=NSLayoutAttributeTrailing,
   width=NSLayoutAttributeWidth,
   height=NSLayoutAttributeHeight,
   centerX=NSLayoutAttributeCenterX,
   centerY=NSLayoutAttributeCenterY,
   baseline=NSLayoutAttributeBaseline,
   lastBaseline=NSLayoutAttributeLastBaseline,
   firstBaseline=NSLayoutAttributeFirstBaseline,
)

_layoutRelationMap = {
    "<=" : NSLayoutRelationLessThanOrEqual,
    "==" : NSLayoutRelationEqual,
    ">=" : NSLayoutRelationGreaterThanOrEqual
}

def _addAutoLayoutRules(obj, rules, metrics=None):
    view = obj._getContentView()
    if metrics is None:
        metrics = {}
    for rule in rules:
        if isinstance(rule, dict):
            view1 = rule["view1"]._getContentView()
            attribute1 = _layoutAttributeMap[rule["attribute1"]]
            relation = _layoutRelationMap[rule.get("relation", "==")]
            view2 = rule["view2"]._getContentView()
            attribute2 = _layoutAttributeMap[rule["attribute2"]]
            multiplier = rule.get("multiplier", 1)
            constant = rule.get("constant", 0)
            constraints = [NSLayoutConstraint.constraintWithItem_attribute_relatedBy_toItem_attribute_multiplier_constant_(
                view1,
                attribute1,
                relation,
                view2,
                attribute2,
                multiplier,
                constant
            )]
        else:
            constraints = NSLayoutConstraint.constraintsWithVisualFormat_options_metrics_views_(
                rule,
                0,
                metrics,
                obj._autoLayoutViews
            )
        view.addConstraints_(constraints)


# --------------------------
# Frame-Based Layout Support
# --------------------------

def _calcFrame(parentFrame, posSize, absolutePositioning=False):
    """
    Convert a vanilla posSize rect to a Cocoa frame.
    """
    (pL, pB), (pW, pH) = parentFrame
    (l, t), (w, h) = posSize
    if not absolutePositioning:
        if l < 0:
            l = pW + l
        if w <= 0:
            w = pW + w - l
        if t < 0:
            t = pH + t
        if h <= 0:
            h = pH + h - t
    b = pH - t - h  # flip it upside down
    return (l, b), (w, h)


def _flipFrame(parentFrame, objFrame):
    """
    Translate a Cocoa frame to vanilla coordinates.
    """
    (pL, pB), (pW, pH) = parentFrame
    (oL, oB), (oW, oH) =  objFrame
    oT = pH - oB - oH
    return oL, oT, oW, oH


# ----------------
# Callback Support
# ----------------

class VanillaCallbackWrapper(NSObject):

    def __new__(cls, callback):
        return cls.alloc().initWithCallback_(callback)

    def initWithCallback_(self, callback):
        self = self.init()
        self.callback = callback
        return self

    def action_(self, sender):
        if hasattr(sender, "vanillaWrapper"):
            sender = sender.vanillaWrapper()
        if self.callback is not None:
            self.callback(sender)


def _breakCycles(view):
    """
    Break cyclic references by deleting _target attributes.
    """
    if hasattr(view, "vanillaWrapper"):
        obj = view.vanillaWrapper()
        if hasattr(obj, "_breakCycles"):
            obj._breakCycles()
    for view in view.subviews():
        _breakCycles(view)



import weakref
from Foundation import NSObject, NSRect, NSMakeRect, NSZeroRect
from AppKit import NSView, NSViewController, NSPopover, NSMinXEdge, NSMaxXEdge, \
    NSMinYEdge, NSMaxYEdge, NSPopoverBehaviorApplicationDefined, NSPopoverBehaviorTransient, \
    NSPopoverBehaviorSemitransient

from vanilla.vanillaBase import VanillaBaseObject, _breakCycles, _addAutoLayoutRules
from vanilla.nsSubclasses import getNSSubclass

_edgeMap = {
    "left": NSMinXEdge,
    "right": NSMaxXEdge,
    "top": NSMinYEdge,
    "bottom": NSMaxYEdge
}


_behaviorMap = {
    "applicationDefined": NSPopoverBehaviorApplicationDefined,
    "transient": NSPopoverBehaviorTransient,
    "semitransient": NSPopoverBehaviorSemitransient
}


class VanillaPopoverContentView(NSView):

    def _getContentView(self):
        return self


class VanillaPopoverDelegate(NSObject):

    def popoverWillShow_(self, notification):
        self.vanillaWrapper()._alertBindings("will show")

    def popoverDidShow_(self, notification):
        self.vanillaWrapper()._alertBindings("did show")

    def popoverWillClose_(self, notification):
        self.vanillaWrapper()._alertBindings("will close")

    def popoverDidClose_(self, notification):
        self.vanillaWrapper()._alertBindings("did close")


class Popover(VanillaBaseObject):

    """
    A popover capable of containing controls.

    .. image:: /_images/Popover.png

    ::

        from vanilla import Window, List, Popover, TextBox

        class PopoverExample:

            def __init__(self):
                self.w = Window((120, 120))
                self.w.list = List((0, 0, -0, -0),
                        ['A', 'B', 'C'],
                        selectionCallback=self.showPopoverCallback)
                self.w.open()

            def showPopoverCallback(self, sender):
                selection = sender.getSelection()
                if not selection:
                    return
                index = sender.getSelection()[0]
                relativeRect = sender.getNSTableView().rectOfRow_(index)
                self.pop = Popover((140, 80))
                self.pop.text = TextBox((10, 10, -10, -10), 'This is a popover.')
                self.pop.open(parentView=sender.getNSTableView(), preferredEdge='right', relativeRect=relativeRect)

        PopoverExample()

    **size** Tuple of form *(width, height)* representing the size of the content
    in the popover.

    **parentView** The parent view that the popover should pop out from. This
    can be either a vanilla object or an instance of `NSView`_ or `NSView_` subclass.

    **preferredEdge** The edge of the parent view that you want the popover
    to pop out from. These are the options:

    +------------+
    | *"left"*   |
    +------------+
    | *"right"*  |
    +------------+
    | *"top"*    |
    +------------+
    | *"bottom"* |
    +------------+

    **behavior** The desired behavior of the popover. These are the options:

    +------------------------+-----------------------------------------------------+
    | *"applicationDefined"* | Corresponds to NSPopoverBehaviorApplicationDefined. |
    +------------------------+-----------------------------------------------------+
    | *"transient"*          | Corresponds to NSPopoverBehaviorTransient.          |
    +------------------------+-----------------------------------------------------+
    | *"semitransient"*      | Corresponds to NSPopoverBehaviorSemitransient.      |
    +------------------------+-----------------------------------------------------+

    .. _NSView: https://developer.apple.com/documentation/appkit/nsview?language=objc
    """

    nsPopoverClass = NSPopover
    contentViewClass = VanillaPopoverContentView
    contentViewControllerClass = NSViewController

    def __init__(self, size, parentView=None, preferredEdge="top", behavior="semitransient"):
        if isinstance(parentView, VanillaBaseObject):
            parentView = parentView._getContentView()
        self._parentView = parentView
        self._preferredEdge = preferredEdge
        # content view and controller
        self._nsObject = getNSSubclass(self.contentViewClass).alloc().initWithFrame_(((0, 0), size))
        self._contentViewController = self.contentViewControllerClass.alloc().init()
        self._contentViewController.setView_(self._nsObject)
        # popover
        cls = getNSSubclass(self.nsPopoverClass)
        self._popover = cls.alloc().init()
        self._popover.setContentViewController_(self._contentViewController)
        self._popover.setBehavior_(_behaviorMap[behavior])
        # delegate
        self._delegate = VanillaPopoverDelegate.alloc().init()
        self._delegate.vanillaWrapper = weakref.ref(self)
        self._popover.setDelegate_(self._delegate)
        self._bindings = {}
        self._autoLayoutViews = {}

    def __del__(self):
        self._breakCycles()

    def _breakCycles(self):
        super(Popover, self)._breakCycles()
        view = self._getContentView()
        if view is not None:
            _breakCycles(view)
        self._contentViewController = None
        self._popover = None
        self._parentView = None
        self._delegate = None

    def getNSPopover(self):
        return self._popover

    def open(self, parentView=None, preferredEdge=None, relativeRect=None):
        """
        Open the popover. If desired, the *parentView* may be specified.
        If not, the values assigned during init will be used.

        Additionally, a rect of form `(x, y, width, height)` may be
        specified to indicate where the popover should pop out from.
        If not provided, the parent view's bounds will be used.
        """
        if isinstance(parentView, VanillaBaseObject):
            parentView = parentView._getContentView()
        if parentView is None:
            parentView = self._parentView
        if relativeRect is not None:
            if not isinstance(relativeRect, NSRect):
                x, y, w, h = relativeRect
                relativeRect = NSMakeRect(x, y, w, h)
        else:
            relativeRect = NSZeroRect
        if preferredEdge is None:
            preferredEdge = self._preferredEdge
        preferredEdge = _edgeMap[preferredEdge]
        self._popover.showRelativeToRect_ofView_preferredEdge_(relativeRect, parentView, preferredEdge)

    def close(self):
        """
        Close the popover.

        Once a popover has been closed it can not be re-opened.
        """
        self._popover.close()

    def resize(self, width, height):
        """
        Change the size of the popover to *width* and *height*.
        """
        self._popover.setContentSize_((width, height))

    def bind(self, event, callback):
        """
        Bind a callback to an event.

        **event** A string representing the desired event. The options are:

        +----------------+-----------------------------------------------+
        | *"will show"*  | Called immediately before the popover shows.  |
        +----------------+-----------------------------------------------+
        | *"did show"*   | Called immediately after the popover shows.   |
        +----------------+-----------------------------------------------+
        | *"will close"* | Called immediately before the popover closes. |
        +----------------+-----------------------------------------------+
        | *"did close"*  | Called immediately after the popover closes.  |
        +----------------+-----------------------------------------------+
        """
        if event not in self._bindings:
            self._bindings[event] = []
        self._bindings[event].append(callback)

    def unbind(self, event, callback):
        """
        Unbind a callback from an event.

        **event** A string representing the desired event.
        Refer to :meth:`Popover.bind` for the options.

        **callback** The callback that has been bound to the event.
        """
        self._bindings[event].remove(callback)

    def _alertBindings(self, key):
        if hasattr(self, "_bindings"):
            if key in self._bindings:
                for callback in self._bindings[key]:
                    # XXX why return? there could be more than one binding.
                    return callback(self)

    def addAutoPosSizeRules(self, rules, metrics=None):
        """
        Add auto layout rules for controls/view in this view.

        **rules** must by a list of rule definitions. Rule definitions may take two forms:

        * strings that follow the `Visual Format Language`_
        * dictionaries with the following key/value pairs:

        +---------------------------+-------------------------------------------------------------------------+
        | key                       | value                                                                   |
        +===========================+=========================================================================+
        | *"view1"*                 | The vanilla wrapped view for the left side of the rule.                 |
        +---------------------------+-------------------------------------------------------------------------+
        | *"attribute1"*            | The attribute of the view for the left side of the rule.                |
        |                           | See below for options.                                                  |
        +---------------------------+-------------------------------------------------------------------------+
        | *"relation"* (optional)   | The relationship between the left side of the rule                      |
        |                           | and the right side of the rule. See below for options.                  |
        |                           | The default value is `"=="`.                                            |
        +---------------------------+-------------------------------------------------------------------------+
        | *"view2"*                 | The vanilla wrapped view for the right side of the rule.                |
        +---------------------------+-------------------------------------------------------------------------+
        | *"attribute2"*            | The attribute of the view for the right side of the rule.               |
        |                           | See below for options.                                                  |
        +---------------------------+-------------------------------------------------------------------------+
        | *"multiplier"* (optional) | The constant multiplied with the attribute on the right side of         |
        |                           | the rule as part of getting the modified attribute.                     |
        |                           | The default value is `1`.                                               |
        +---------------------------+-------------------------------------------------------------------------+
        | *"constant"* (optional)   | The constant added to the multiplied attribute value on the right       |
        |                           | side of the rule to yield the final modified attribute.                 |
        |                           | The default value is `0`.                                               |
        +---------------------------+-------------------------------------------------------------------------+

        The `attribute1` and `attribute2` options are:

        +-------------------+--------------------------------+
        | value             | AppKit equivalent              |
        +===================+================================+
        | *"left"*          | NSLayoutAttributeLeft          |
        +-------------------+--------------------------------+
        | *"right"*         | NSLayoutAttributeRight         |
        +-------------------+--------------------------------+
        | *"top"*           | NSLayoutAttributeTop           |
        +-------------------+--------------------------------+
        | *"bottom"*        | NSLayoutAttributeBottom        |
        +-------------------+--------------------------------+
        | *"leading"*       | NSLayoutAttributeLeading       |
        +-------------------+--------------------------------+
        | *"trailing"*      | NSLayoutAttributeTrailing      |
        +-------------------+--------------------------------+
        | *"width"*         | NSLayoutAttributeWidth         |
        +-------------------+--------------------------------+
        | *"height"*        | NSLayoutAttributeHeight        |
        +-------------------+--------------------------------+
        | *"centerX"*       | NSLayoutAttributeCenterX       |
        +-------------------+--------------------------------+
        | *"centerY"*       | NSLayoutAttributeCenterY       |
        +-------------------+--------------------------------+
        | *"baseline"*      | NSLayoutAttributeBaseline      |
        +-------------------+--------------------------------+
        | *"lastBaseline"*  | NSLayoutAttributeLastBaseline  |
        +-------------------+--------------------------------+
        | *"firstBaseline"* | NSLayoutAttributeFirstBaseline |
        +-------------------+--------------------------------+

        Refer to the `NSLayoutAttribute documentation`_ for the information about what each of these do.

        The `relation` options are:

        +--------+------------------------------------+
        | value  | AppKit equivalent                  |
        +========+====================================+
        | *"<="* | NSLayoutRelationLessThanOrEqual    |
        +--------+------------------------------------+
        | *"=="* | NSLayoutRelationEqual              |
        +--------+------------------------------------+
        | *">="* | NSLayoutRelationGreaterThanOrEqual |
        +--------+------------------------------------+

        Refer to the `NSLayoutRelation documentation`_ for the information about what each of these do.

        **metrics** may be either *None* or a dict containing key value pairs
        representing metrics keywords used in the rules defined with strings.

        .. _Visual Format Language: https://developer.apple.com/library/archive/documentation/UserExperience/Conceptual/AutolayoutPG/VisualFormatLanguage.html#//apple_ref/doc/uid/TP40010853-CH27-SW1
        .. _NSLayoutAttribute documentation: https://developer.apple.com/documentation/uikit/nslayoutattribute?language=objc
        .. _NSLayoutRelation documentation: https://developer.apple.com/documentation/uikit/nslayoutrelation?language=objc

        """
        _addAutoLayoutRules(self, rules, metrics)



from Foundation import NSObject
from AppKit import NSTextView, NSScrollView, NSBezelBorder, NSViewWidthSizable, NSViewHeightSizable
from vanilla.nsSubclasses import getNSSubclass
from vanilla.vanillaBase import VanillaBaseObject, VanillaCallbackWrapper


class VanillaTextEditorDelegate(NSObject):

    def textDidChange_(self, notification):
        if hasattr(self, "_target"):
            textView = notification.object()
            self._target.action_(textView)


class TextEditor(VanillaBaseObject):

    """
    Standard long text entry control.

    .. image:: /_images/TextEditor.png

    ::

        from vanilla import Window, TextEditor

        class TextEditorDemo:

            def __init__(self):
                self.w = Window((200, 200))
                self.w.textEditor = TextEditor((10, 10, -10, -10),
                                    text='hello world',
                                    callback=self.textEditorCallback)
                self.w.open()

            def textEditorCallback(self, sender):
                print("text entry!", sender.get())

        TextEditorDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the text entry control.

    **text** The text to be displayed in the text entry control.

    **callback** The method to be called when the user presses the text
    entry control.

    **readOnly** Boolean representing if the text can be edited or not.

    **checksSpelling** Boolean representing if spelling should be automatically
    checked or not.
    """

    nsScrollViewClass = NSScrollView
    nsTextViewClass = NSTextView
    delegateClass = VanillaTextEditorDelegate

    def __init__(self, posSize, text="", callback=None, readOnly=False, checksSpelling=False):
        self._posSize = posSize
        self._nsObject = self.nsScrollViewClass.alloc().init()  # no need to do getNSSubclass() here
        self._nsObject.setHasVerticalScroller_(True)
        self._nsObject.setBorderType_(NSBezelBorder)
        self._nsObject.setDrawsBackground_(True)
        self._textView = getNSSubclass(self.nsTextViewClass)(self)
        self._textView.setAllowsUndo_(True)
        self._textView.setString_(text)
        self._textView.setContinuousSpellCheckingEnabled_(checksSpelling)
        self._textView.setAutoresizingMask_(NSViewWidthSizable | NSViewHeightSizable)
        self._textView.setEditable_(not readOnly)
        self._nsObject.setDocumentView_(self._textView)
        # do the base object init methods
        self._setCallback(callback)
        self._setAutosizingFromPosSize(posSize)

    def _testForDeprecatedAttributes(self):
        super(TextEditor, self)._testForDeprecatedAttributes()
        from warnings import warn
        if hasattr(self, "_textViewClass"):
            warn(DeprecationWarning("The _textViewClass attribute is deprecated. Use the nsTextViewClass attribute."))
            self.nsTextViewClass = self._textViewClass

    def getNSScrollView(self):
        """
        Return the `NSScrollView`_ that this object wraps.

        .. _NSScrollView: https://developer.apple.com/documentation/appkit/nsscrollview?language=objc
        """
        return self._nsObject

    def getNSTextView(self):
        """
        Return the `NSTextView`_ that this object wraps.

        .. _NSTextView: https://developer.apple.com/documentation/appkit/nstextview?language=objc
        """
        return self._textView

    def _setCallback(self, callback):
        if callback is not None:
            self._target = VanillaCallbackWrapper(callback)
            delegate = self._textView.delegate()
            if delegate is None:
                self._textViewDelegate = delegate = self.delegateClass.alloc().init()
                self._textView.setDelegate_(delegate)
            delegate._target = self._target

    def get(self):
        """
        Get the contents of the text entry control.
        """
        return str(self._textView.string())

    def set(self, value):
        """
        Set the contents of the text box.

        **value** A string representing the contents of the text box.
        """
        self._textView.setString_(value)

    def selectAll(self):
        """
        Select all text in the text entry control.
        """
        self._textView.selectAll_(None)

    #def selectLine(self, lineNumber, charOffset=0):
    #    raise NotImplementedError
    #
    #def getSelection(self):
    #    """
    #    Get the selected text.
    #    """
    #    selStart, selEnd = self._textView.selectedRange()
    #    return selStart, selStart+selEnd
    #
    #def setSelection(self, selStart, selEnd):
    #    selEnd = selEnd - selStart
    #    self._textView.setSelectedRange_((selStart, selEnd))
    #
    #def getSelectedText(self):
    #    selStart, selEnd = self.getSelection()
    #    return self._textView.string()[selStart:selEnd]
    #
    #def expandSelection(self):
    #    raise NotImplementedError
    #
    #def insert(self, text):
    #    self._textView.insert_(text)



from vanilla import *


class TestSplitView(object):

    def __init__(self, view):

        self.w = Window((400, 400), "%s" % view.__name__, minSize=(200, 200))

        anEditor = TextEditor((0, 0, -0, -0))
        aList = List((0, 0, -0, -0), ["a", "b", "c"])

        group = Group((0, 0, -0, -0))
        group.aList = List((0, 0, -0, -0), ["x", "y", "z"])

        nestedPaneDescriptions = [
            dict(view=group, identifier="group", canCollapse=False),
            dict(view=aList, identifier="aList", minSize=100, maxSize=300)
        ]

        nestedSplit = view((0, 0, -0, -0), paneDescriptions=nestedPaneDescriptions)

        paneDescriptions = [
            dict(view=anEditor, identifier="anEditor", canCollapse=False),
            dict(view=nestedSplit, identifier="nestedSplit", minSize=100, maxSize=300),
        ]
        self.w.splitView = view((0, 0, -0, -0), paneDescriptions=paneDescriptions, isVertical=False)
        self.w.open()


def testSplitView():

    for view in SplitView, SplitView2:
        TestSplitView(view)


testSplitView()



from AppKit import NSObject, NSComboBox
from objc import python_method
from vanilla.vanillaBase import VanillaBaseControl


class VanillaComboBoxDelegate(NSObject):

    _continuous = True

    @python_method
    def _callVanillaCallback(self, notification):
        obj = notification.object()
        target = obj.target()
        if target is not None:
            target.action_(obj)

    def controlTextDidChange_(self, notification):
        if self._continuous:
            self._callVanillaCallback(notification)

    def controlTextDidEndEditing_(self, notification):
        if not self._continuous:
            self._callVanillaCallback(notification)

    def comboBoxSelectionDidChange_(self, notification):
        obj = notification.object()
        obj.setObjectValue_(obj.objectValueOfSelectedItem())
        self._callVanillaCallback(notification)


class ComboBox(VanillaBaseControl):

    """
    A text entry control that allows direct text entry or selection for a list
    of options.

    .. image:: /_images/ComboBox.png

    ::

        from vanilla import Window, ComboBox

        class ComboBoxDemo:

            def __init__(self):
                self.w = Window((100, 41))
                self.w.comboBox = ComboBox((10, 10, -10, 21),
                                    ["AA", "BB", "CC", "DD"],
                                    callback=self.comboBoxCallback)
                self.w.open()

            def comboBoxCallback(self, sender):
                print("combo box entry!", sender.get())

        ComboBoxDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the combo box control. The size of
    the combo box should match the appropriate value for the given *sizeStyle*.

    +---------------------------+
    | **Standard Dimensions**   |
    +-----------+---+-----------+
    | *Regular* | H | 21        |
    +-----------+---+-----------+
    | *Small*   | H | 17        |
    +-----------+---+-----------+
    | *Mini*    | H | 14        |
    +-----------+---+-----------+

    **items** The items to be displayed in the combo box.

    **completes** Boolean representing if the combo box auto completes
    entered text.

    **continuous** If *True*, the callback (if any) will be called upon each
    keystroke, if *False*, only call the callback when editing finishes or
    after item selection. Default is *False*.

    **callback** The method to be called when the user enters text.

    **formatter** A `NSFormatter`_ for controlling the display and input of
    the text entry.

    **sizeStyle** A string representing the desired size style of the combo box.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+

    .. _NSFormatter: https://developer.apple.com/documentation/foundation/nsformatter?language=objc
    """

    nsComboBoxClass = NSComboBox

    frameAdjustments = {
        "mini": (0, -4, 1, 5),
        "small": (0, -4, 3, 5),
        "regular": (0, -3, 3, 5),
    }

    def __init__(self, posSize, items, completes=True, continuous=False,
            callback=None, formatter=None, sizeStyle="regular"):
        self._continuous = continuous
        self._setupView(self.nsComboBoxClass, posSize, callback=callback)
        self._setSizeStyle(sizeStyle)
        self._nsObject.addItemsWithObjectValues_(items)
        self._nsObject.setCompletes_(completes)
        if formatter is not None:
            self._nsObject.setFormatter_(formatter)

    def _breakCycles(self):
        super(ComboBox, self)._breakCycles()
        self._delegate = None

    def _setCallback(self, callback):
        super(ComboBox, self)._setCallback(callback)
        if callback is not None:
            self._delegate = VanillaComboBoxDelegate.alloc().init()
            self._delegate._continuous = self._continuous
            self._nsObject.setDelegate_(self._delegate)

    def getNSComboBox(self):
        """
        Return the `NSComboBox`_ that this object wraps.

        .. _NSComboBox: https://developer.apple.com/documentation/appkit/nscombobox?language=objc
        """
        return self._nsObject

    def get(self):
        """
        Get the text entered in the combo box.
        """
        return self._nsObject.objectValue()

    def set(self, value):
        """
        Set the text in the text field of the combo box.

        **value** A string to set in the combo box.
        """
        self._nsObject.setObjectValue_(value)

    def setItems(self, items):
        """
        Set the items in the combo box list.

        **items** A list of strings to set in the combo box list.
        """
        self._nsObject.removeAllItems()
        self._nsObject.addItemsWithObjectValues_(items)

    def getItems(self):
        """
        Get the items in the combo box as a list.
        """
        return list(self._nsObject.objectValues())



import objc
import weakref


class _VanillaMethods:

    def __new__(cls, wrapper):
        self = cls.alloc().init()
        self.setVanillaWrapper_(wrapper)
        return self

    def setVanillaWrapper_(self, wrapper):
        self._wrapperRef = weakref.ref(wrapper)
        return self

    def vanillaWrapper(self):
        try:
            return self._wrapperRef()
        except AttributeError:
            return None


_subclasses = {}
def getNSSubclass(classOrName=None):
    """
    Return a subclass of a given Objective-C class.
    This subclass allows us to store a weakref to a Vanilla wrapper
    object. This way we can always get back at it. Mainly meant to
    translate "sender" arguments of action methods to a Vanilla object.

    The returned subclass can be instantiated from Python like this:

        >>> class Wrapper(object): pass
        ...
        >>> wrpr = Wrapper()
        >>> view = getNSSubclass("NSBox")(wrpr)
        >>> view.vanillaWrapper() is wrpr
        True

        >>> from Foundation import NSObject
        >>> class MyCustomNSClass(NSObject): pass
        ...
        >>> view = getNSSubclass(MyCustomNSClass)(wrpr)
        >>> view.vanillaWrapper() is wrpr
        True

    """
    subCls = None
    if isinstance(classOrName, str):
        cls = objc.lookUpClass(classOrName)
        className = classOrName
    else:
        cls = classOrName
        className = cls.__name__
    subCls = _subclasses.get(className)
    if subCls is None:
        vName = "V" + className
        subCls = cls.__class__(vName, (cls,), _VanillaMethods.__dict__.copy())
        _subclasses[className] = subCls
    return subCls


def _doctest():
    """
        >>> from AppKit import NSTextView
        >>> sc = getNSSubclass('NSTextView')
        >>> sc.__name__
        'VNSTextView'
        >>> issubclass(sc, NSTextView)
        True
        >>> hasattr(sc, 'setVanillaWrapper_')
        True
        >>> hasattr(sc, 'vanillaWrapper')
        True
        >>> # duplicate test to test caching
        >>> sc = getNSSubclass('NSTextView')
        >>> sc.__name__
        'VNSTextView'
        >>> issubclass(sc, NSTextView)
        True
        >>> class Wrapper(object): pass
        ...
        >>> wrpr = Wrapper()
        >>> isinstance(sc(wrpr), NSTextView)
        True
        >>> # test custom class
        >>> from Foundation import NSObject
        >>> class TestClass(NSObject): pass
        ...
        >>> sc = getNSSubclass(TestClass)
        >>> sc.__name__
        'VTestClass'
        >>> isinstance(sc(wrpr), NSObject)
        True
        >>> isinstance(sc(wrpr), TestClass)
        True
        >>> hasattr(sc, 'setVanillaWrapper_')
        True
        >>> hasattr(sc, 'vanillaWrapper')
        True
    """

if __name__ == "__main__":
    import doctest
    doctest.testmod()



from Foundation import NSObject
from AppKit import NSTabView, NSTabViewItem, NSNoTabsNoBorder, NSFont
from vanilla.vanillaBase import VanillaBaseObject, _breakCycles, _sizeStyleMap, VanillaCallbackWrapper, \
        _reverseSizeStyleMap


class VanillaTabItem(VanillaBaseObject):

    nsTabViewItemClass = NSTabViewItem

    def __init__(self, title):
        self._autoLayoutViews = {}
        self._tabItem = self.nsTabViewItemClass.alloc().initWithIdentifier_(title)
        self._tabItem.setLabel_(title)

    def _getContentView(self):
        return self._tabItem.view()

    def _breakCycles(self):
        _breakCycles(self._tabItem.view())
        self._autoLayoutViews.clear()


class VanillaTabsDelegate(NSObject):

    def tabView_didSelectTabViewItem_(self, tabView, tabViewItem):
        if hasattr(self, "_target"):
            self._target.action_(tabView.vanillaWrapper())


class Tabs(VanillaBaseObject):

    """
    A set of tabs attached to a window. Each tab is capable of containing controls.

    .. image:: /_images/Tabs.png

    To add a control to a tab, simply set it as an attribute of the tab.

    ::

        from vanilla import Window, Tabs, TextBox

        class TabsDemo:

            def __init__(self):
                self.w = Window((250, 100))
                self.w.tabs = Tabs((10, 10, -10, -10), ["Tab One", "Tab Two"])
                tab1 = self.w.tabs[0]
                tab1.text = TextBox((10, 10, -10, -10), "This is tab 1")
                tab2 = self.w.tabs[1]
                tab2.text = TextBox((10, 10, -10, -10), "This is tab 2")
                self.w.open()

        TabsDemo()

    No special naming is required for the attributes. However, each attribute
    must have a unique name.

    To retrieve a particular tab, access it by index::

        myTab = self.w.tabs[0]

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing the position
    and size of the tabs.

    **titles** An ordered list of tab titles.

    **callback** The method to be called when the user selects a new tab.

    **sizeStyle** A string representing the desired size style of the tabs.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+
    """

    nsTabViewClass = NSTabView
    vanillaTabViewItemClass = VanillaTabItem

    allFrameAdjustments = {
        # The sizeStyle will be part of the
        # className used for the lookup here.
        "Tabs-mini": (-7, -10, 14, 12),
        "Tabs-small": (-7, -10, 14, 13),
        "Tabs-regular": (-7, -10, 14, 16),
    }

    def __init__(self, posSize, titles=["Tab"], callback=None, sizeStyle="regular", showTabs=True):
        self._setupView(self.nsTabViewClass, posSize) # hold off on setting callback
        self._setSizeStyle(sizeStyle)
        self._tabItems = []
        for title in titles:
            tab = self.vanillaTabViewItemClass(title)
            self._tabItems.append(tab)
            self._nsObject.addTabViewItem_(tab._tabItem)
        if not showTabs:
            self._nsObject.setTabViewType_(NSNoTabsNoBorder)
            self._nsObject.setDrawsBackground_(False)
        # now that the tabs are all set, set the callback.
        # this is done because the callback will be called
        # while the tabs are being added.
        if callback is not None:
            self._setCallback(callback)

    def getNSTabView(self):
        """
        Return the `NSTabView`_ that this object wraps.

        .. _NSTabView: https://developer.apple.com/documentation/appkit/nstabview?language=objc
        """
        return self._nsObject

    def _adjustPosSize(self, frame):
        if self._nsObject.tabViewType() == NSNoTabsNoBorder:
            return frame
        sizeStyle = _reverseSizeStyleMap[self._nsObject.controlSize()]
        tabsType = "Tabs-" + sizeStyle
        self.frameAdjustments = self.allFrameAdjustments[tabsType]
        return super(Tabs, self)._adjustPosSize(frame)

    def _setCallback(self, callback):
        if callback is not None:
            self._target = VanillaCallbackWrapper(callback)
            delegate = self._nsObject.delegate()
            if delegate is None:
                self._delegate = delegate = VanillaTabsDelegate.alloc().init()
                self._nsObject.setDelegate_(delegate)
            delegate._target = self._target

    def _setSizeStyle(self, value):
        value = _sizeStyleMap[value]
        self._nsObject.setControlSize_(value)
        font = NSFont.systemFontOfSize_(NSFont.systemFontSizeForControlSize_(value))
        self._nsObject.setFont_(font)

    def __getitem__(self, index):
        return self._tabItems[index]

    def _breakCycles(self):
        super(Tabs, self)._breakCycles()
        for item in self._tabItems:
            item._breakCycles()

    def get(self):
        """
        Get the index of the selected tab.
        """
        item = self._nsObject.selectedTabViewItem()
        index = self._nsObject.indexOfTabViewItem_(item)
        return index

    def set(self, value):
        """
        Set the selected tab.

        **value** The index of the tab to be selected.
        """
        self._nsObject.selectTabViewItemAtIndex_(value)



from AppKit import NSSearchField
from vanilla.vanillaBase import VanillaBaseControl


class SearchBox(VanillaBaseControl):

    """
    A text entry field similar to the search field in Safari.

    .. image:: /_images/SearchBox.png

    ::

        from vanilla import Window, SearchBox

        class SearchBoxDemo:

            def __init__(self):
                self.w = Window((120, 42))
                self.w.searchBox = SearchBox((10, 10, -10, 22),
                                        callback=self.searchBoxCallback)
                self.w.open()

            def searchBoxCallback(self, sender):
                print("search box entry!", sender.get())

        SearchBoxDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the search box.

    +-------------------------+
    | **Standard Dimensions** |
    +---------+---+-----------+
    | Regular | H | 22        |
    +---------+---+-----------+
    | Small   | H | 19        |
    +---------+---+-----------+
    | Mini    | H | 15        |
    +---------+---+-----------+

    **text** The text to be displayed in the search box.

    **callback** The method to be called when the user presses the search box.

    **formatter** A `NSFormatter`_ for controlling the display and input of
    the text entry.

    **placeholder** A placeholder string to be shown when the text entry
    control is empty.

    **sizeStyle** A string representing the desired size style of the
    search box. The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+

    .. _NSFormatter: https://developer.apple.com/documentation/foundation/nsformatter?language=objc
    """

    nsSearchFieldClass = NSSearchField

    def __init__(self, posSize, text="", callback=None, formatter=None, placeholder=None, sizeStyle="regular"):
        self._setupView(self.nsSearchFieldClass, posSize, callback=callback)
        self._setSizeStyle(sizeStyle)
        self._nsObject.setStringValue_(text)
        if formatter is not None:
            self._nsObject.setFormatter_(formatter)
        cell = self._nsObject.cell()
        cell.setScrollable_(True)
        if placeholder:
            cell.setPlaceholderString_(placeholder)

    def getNSSearchField(self):
        """
        Return the `NSSearchField`_ that this object wraps.

        .. _NSSearchField: https://developer.apple.com/documentation/appkit/nssearchfield?language=objc
        """
        return self._nsObject

    def get(self):
        """
        Get the contents of the search box.
        """
        return self._nsObject.stringValue()

    def set(self, value):
        """
        Set the contents of the search box.

        **value** A string representing the contents of the search box.
        """
        self._nsObject.setStringValue_(value)



from AppKit import NSMatrix, NSRadioButton, NSButtonCell, NSRadioModeMatrix, NSFont, NSCellDisabled
from vanilla.vanillaBase import VanillaBaseControl, _sizeStyleMap
from vanilla.vanillaButton import Button
from vanilla.vanillaStackGroup import VerticalStackGroup, HorizontalStackGroup

try:
    from AppKit import NSRadioButtonType
except ImportError:
    from AppKit import NSRadioButton
    NSRadioButtonType = NSRadioButton


class _RadioGroupMixin(object):

    _heights = dict(
        regular=18,
        small=15,
        mini=12
    )

    def _init(self, cls, posSize, titles, callback=None, sizeStyle="regular"):
        spacing = self._spacing[sizeStyle]
        self._buttonHeight = self._heights[sizeStyle]
        self._callback = callback
        self._sizeStyle = sizeStyle
        super(cls, self).__init__(posSize, spacing=spacing, alignment="leading")
        self._buildButtons(titles, sizeStyle)

    def _buildButtons(self, titles, sizeStyle):
        self._buttons = []
        for title in titles:
            button = RadioButton("auto", title, callback=self._buttonCallback, sizeStyle=sizeStyle)
            self._buttons.append(button)
            self.addView(button, height=self._buttonHeight)

    def _buttonCallback(self, sender):
        for button in self._buttons:
            if button != sender:
                button.set(False)
        if self._callback is not None:
            self._callback(self)

    def getFittingHeight(self):
        """
        Get the fitting height for all buttons in the group.
        """
        count = len(self._buttons)
        size = self._heights[self._sizeStyle]
        spacing = self._spacing[self._sizeStyle]
        height = size * count
        height += spacing * (count - 1)
        return height

    def get(self):
        """
        Get the index of the selected radio button.
        """
        for index, button in enumerate(self._buttons):
            if button.get():
                return index

    def set(self, index):
        """
        Set the index of the selected radio button.
        """
        for other, button in enumerate(self._buttons):
            button.set(other == index)


class VerticalRadioGroup(VerticalStackGroup, _RadioGroupMixin):

    """
    A vertical collection of radio buttons.

    .. image:: /_images/VerticalRadioGroup.png

    ::

        from vanilla import Window, VerticalRadioGroup

        class VerticalRadioGroupDemo:

            def __init__(self):
                self.w = Window((100, 100))
                self.w.radioGroup = VerticalRadioGroup(
                    "auto",
                    ["Option 1", "Option 2"],
                    callback=self.radioGroupCallback
                )
                self.w.radioGroup.set(0)
                rules = [
                    "H:|-[radioGroup]-|",
                    "V:|-[radioGroup(==%d)]-|" % self.w.radioGroup.getFittingHeight()
                ]
                self.w.addAutoPosSizeRules(rules)
                self.w.open()

            def radioGroupCallback(self, sender):
                print("radio group edit!", sender.get())

        VerticalRadioGroupDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the radio group.

    **titles** A list of titles to be shown next to the radio buttons.

    **callback** The method to be caled when a radio button is selected.

    **sizeStyle** A string representing the desired size style of the radio group.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+
    """

    _spacing = dict(
        regular=2,
        small=2,
        mini=2
    )

    def __init__(self, posSize, titles, callback=None, sizeStyle="regular"):
        self._init(VerticalRadioGroup, posSize, titles, callback=callback, sizeStyle=sizeStyle)


class HorizontalRadioGroup(HorizontalStackGroup, _RadioGroupMixin):

    ### TODO: Example is not horizontal but vertical!

    """
    A horizontal collection of radio buttons.

    ::

        from vanilla import Window, HorizontalRadioGroup

        class RadioGroupDemo:

            def __init__(self):
                self.w = Window((300, 100))
                self.w.radioGroup = HorizontalRadioGroup(
                    "auto",
                    ["Option 1", "Option 2"],
                    callback=self.radioGroupCallback
                )
                self.w.radioGroup.set(0)
                rules = [
                    "H:|-[radioGroup]-|",
                    "V:|-[radioGroup(==%d)]-|" % self.w.radioGroup.getFittingHeight()
                ]
                self.w.addAutoPosSizeRules(rules)
                self.w.open()

            def radioGroupCallback(self, sender):
                print("radio group edit!", sender.get())

        RadioGroupDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the radio group.

    **titles** A list of titles to be shown next to the radio buttons.

    **callback** The method to be caled when a radio button is selected.

    **sizeStyle** A string representing the desired size style of the radio group.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+
    """

    _spacing = dict(
        regular=4,
        small=3,
        mini=3
    )

    def __init__(self, posSize, titles, callback=None, sizeStyle="regular"):
        self._init(HorizontalRadioGroup, posSize, titles, callback=callback, sizeStyle=sizeStyle)


class RadioButton(Button):

    """
    A single radio button.
    """

    nsBezelStyle = None
    nsButtonType = NSRadioButtonType

    def __init__(self, posSize, title, value=False, callback=None, sizeStyle="regular"):
        super(RadioButton, self).__init__(posSize, title, callback=callback, sizeStyle=sizeStyle)
        self.set(value)

    def set(self, value):
        """
        Set the state of the radio button.

        **value** A boolean representing the state of the radio button.
        """
        self._nsObject.setState_(value)

    def get(self):
        """
        Get the state of the radio button.
        """
        return self._nsObject.state()


# ------
# Legacy
# ------

class RadioGroup(VanillaBaseControl):

    """
    A collection of radio buttons.

    .. image:: /_images/RadioGroup.png

    .. note:: This should be used only for frame layout.

    ::

        from vanilla import Window, RadioGroup

        class RadioGroupDemo:

            def __init__(self):
                self.w = Window((100, 60))
                self.w.radioGroup = RadioGroup((10, 10, -10, 40),
                                        ["Option 1", "Option 2"],
                                        callback=self.radioGroupCallback)
                self.w.radioGroup.set(0)
                self.w.open()

            def radioGroupCallback(self, sender):
                print("radio group edit!", sender.get())

        RadioGroupDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the radio group.

    **titles** A list of titles to be shown next to the radio buttons.

    **isVertical** Boolean representing if the radio group is
    vertical or horizontal.

    **callback** The method to be caled when a radio button is selected.

    **sizeStyle** A string representing the desired size style of the radio group.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+
    """

    nsMatrixClass = NSMatrix
    nsCellClass = NSButtonCell

    def __init__(self, posSize, titles, isVertical=True, callback=None, sizeStyle="regular"):
        self._setupView(self.nsMatrixClass, posSize, callback=callback)
        self._isVertical = isVertical
        matrix = self._nsObject
        matrix.setMode_(NSRadioModeMatrix)
        matrix.setCellClass_(self.nsCellClass)
        # XXX! this does not work for vertical radio groups!
        matrix.setAutosizesCells_(True)
        # we handle the control size setting here
        # since the actual NS object is a NSMatrix
        cellSizeStyle = _sizeStyleMap[sizeStyle]
        font = NSFont.systemFontOfSize_(NSFont.systemFontSizeForControlSize_(cellSizeStyle))
        # intercell spacing and cell spacing are based on the sizeStyle
        if posSize == "auto":
            w = 0
        else:
            w = posSize[2]
        if sizeStyle == "regular":
            matrix.setIntercellSpacing_((4.0, 2.0))
            matrix.setCellSize_((w, 18))
        elif sizeStyle == "small":
            matrix.setIntercellSpacing_((3.5, 2.0))
            matrix.setCellSize_((w, 15))
        elif sizeStyle == "mini":
            matrix.setIntercellSpacing_((3.0, 2.0))
            matrix.setCellSize_((w, 12))
        else:
            raise ValueError("sizeStyle must be 'regular', 'small' or 'mini'")
        for _ in range(len(titles)):
            if isVertical:
                matrix.addRow()
            else:
                matrix.addColumn()
        for title, cell in zip(titles, matrix.cells()):
            cell.setButtonType_(NSRadioButton)
            cell.setTitle_(title)
            cell.setControlSize_(cellSizeStyle)
            cell.setFont_(font)

    def _testForDeprecatedAttributes(self):
        super(RadioGroup, self)._testForDeprecatedAttributes()
        from warnings import warn
        if hasattr(self, "_cellClass"):
            warn(DeprecationWarning("The _cellClass attribute is deprecated. Use the nsCellClass attribute."))
            self.nsCellClass = self._cellClass

    def getNSMatrix(self):
        """
        Return the `NSMatrix`_ that this object wraps.

        .. _NSMatrix: https://developer.apple.com/documentation/appkit/nsmatrix?language=objc
        """
        return self._nsObject

    def get(self):
        """
        Get the index of the selected radio button.
        """
        if self._isVertical:
            return self._nsObject.selectedRow()
        else:
            return self._nsObject.selectedColumn()

    def set(self, index):
        """
        Set the index of the selected radio button.
        """
        if self._isVertical:
            row = index
            column = 0
        else:
            row = 0
            column = index
        self._nsObject.selectCellAtRow_column_(row, column)

    def enableRadioButton(self, index, onOff=True):
        """
        Enable or disable a RadioGroup button specified by its index.
        """
        self._nsObject.cells()[index].setCellAttribute_to_(NSCellDisabled, not onOff)



from AppKit import NSImageView, NSImage, NSImageAlignCenter, NSImageAlignLeft, NSImageAlignRight, NSImageAlignTop, NSImageAlignTopLeft, NSImageAlignTopRight, NSImageAlignBottom, NSImageAlignBottomLeft, NSImageAlignBottomRight, NSScaleProportionally, NSScaleNone, NSScaleToFit
from vanilla.vanillaBase import VanillaBaseObject

_imageAlignmentMap = {
    ("center", "center") : NSImageAlignCenter,
    ("left", "center") : NSImageAlignLeft,
    ("right", "center") : NSImageAlignRight,
    ("center", "top") : NSImageAlignTop,
    ("left", "top") : NSImageAlignTopLeft,
    ("right", "top") : NSImageAlignTopRight,
    ("center", "bottom") : NSImageAlignBottom,
    ("left", "bottom") : NSImageAlignBottomLeft,
    ("right", "bottom") : NSImageAlignBottomRight
}

_imageScaleMap = {
    "proportional" : NSScaleProportionally,
    "none" : NSScaleNone,
    "fit" : NSScaleToFit
}


class ImageView(VanillaBaseObject):

    """
    A view that displays an image.

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the view.

    **horizontalAlignment** A string representing the desired horizontal
    alignment of the image in the view. The options are:

    +-------------+-------------------------+
    | "left"      | Image is aligned left.  |
    +-------------+-------------------------+
    | "right"     | Image is aligned right. |
    +-------------+-------------------------+
    | "center"    | Image is centered.      |
    +-------------+-------------------------+

    **verticalAlignment** A string representing the desired vertical alignment
    of the image in the view. The options are:

    +-------------+--------------------------+
    | "top"       | Image is aligned top.    |
    +-------------+--------------------------+
    | "bottom"    | Image is aligned bottom. |
    +-------------+--------------------------+
    | "center"    | Image is centered.       |
    +-------------+--------------------------+

    **scale** A string representing the desired scale style of the image in the
    view. The options are:

    +----------------+----------------------------------------------+
    | "proportional" | Proportionally scale the image to fit in the |
    |                | view if it is larger than the view.          |
    +----------------+----------------------------------------------+
    | "fit"          | Distort the proportions of the image until   |
    |                | it fits exactly in the view.                 |
    +----------------+----------------------------------------------+
    | "none"         | Do not scale the image.                      |
    +----------------+----------------------------------------------+
    """

    nsImageViewClass = NSImageView

    def __init__(self, posSize, horizontalAlignment="center", verticalAlignment="center", scale="proportional"):
        self._setupView(self.nsImageViewClass, posSize)
        align = _imageAlignmentMap[(horizontalAlignment, verticalAlignment)]
        self._nsObject.setImageAlignment_(align)
        scale = _imageScaleMap[scale]
        self._nsObject.setImageScaling_(scale)

    def getNSImageView(self):
        """
        Return the `NSImageView`_ that this object wraps.

        .. _NSImageView: https://developer.apple.com/documentation/appkit/nsimageview?language=objc
        """
        return self._nsObject

    def setImage(self, imagePath=None, imageNamed=None, imageObject=None):
        """
        Set the image in the view.

        **imagePath** A file path to an image.

        **imageNamed** The name of an image already load as a `NSImage`_
        by the application.

        **imageObject** A `NSImage`_ object.

        .. note::
           Only one of *imagePath*, *imageNamed*, *imageObject* should be set.

        .. _NSImage: https://developer.apple.com/documentation/appkit/nsimage?language=objc
        """
        if imagePath is not None:
            image = NSImage.alloc().initWithContentsOfFile_(imagePath)
        elif imageNamed is not None:
            image = NSImage.imageNamed_(imageNamed)
        elif imageObject is not None:
            image = imageObject
        else:
            raise ValueError("no image source defined")
        self._nsObject.setImage_(image)



from AppKit import NSButton, NSImage, NSBundle, NSRoundedBezelStyle, NSShadowlessSquareBezelStyle, NSHelpButtonBezelStyle, NSMomentaryPushInButton, NSCommandKeyMask, NSAlternateKeyMask, NSAlternateKeyMask, NSShiftKeyMask, NSAlphaShiftKeyMask, NSHelpFunctionKey, NSHomeFunctionKey, NSEndFunctionKey, NSPageUpFunctionKey, NSPageDownFunctionKey, NSDeleteFunctionKey, NSLeftArrowFunctionKey, NSRightArrowFunctionKey, NSUpArrowFunctionKey, NSDownArrowFunctionKey, NSImageLeft, NSImageRight, NSImageAbove, NSImageBelow, NSImageOnly, NSTextAlignmentRight, NSTextAlignmentLeft, NSNoCellMask
from vanilla.vanillaBase import VanillaBaseControl


_modifierMap = {
    "command": NSCommandKeyMask,
    "control": NSAlternateKeyMask,
    "option": NSAlternateKeyMask,
    "shift": NSShiftKeyMask,
    "capslock": NSAlphaShiftKeyMask,
}

_keyMap = {
    "help": NSHelpFunctionKey,
    "home": NSHomeFunctionKey,
    "end": NSEndFunctionKey,
    "pageup": NSPageUpFunctionKey,
    "pagedown": NSPageDownFunctionKey,
    "forwarddelete": NSDeleteFunctionKey,
    "leftarrow": NSLeftArrowFunctionKey,
    "rightarrow": NSRightArrowFunctionKey,
    "uparrow": NSUpArrowFunctionKey,
    "downarrow": NSDownArrowFunctionKey,
}


class Button(VanillaBaseControl):

    """
    A standard button.

    .. image:: /_images/Button.png

    ::

        from vanilla import Window, Button

        class ButtonDemo:

             def __init__(self):
                 self.w = Window((100, 40))
                 self.w.button = Button((10, 10, -10, 20), "A Button",
                                    callback=self.buttonCallback)
                 self.w.open()

             def buttonCallback(self, sender):
                 print("button hit!")

        ButtonDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the button. The size of the button
    should match the appropriate value for the given *sizeStyle*.

    +-------------------------+
    | **Standard Dimensions** |
    +=========+===+===========+
    | Regular | H | 20        |
    +---------+---+-----------+
    | Small   | H | 17        |
    +---------+---+-----------+
    | Mini    | H | 14        |
    +---------+---+-----------+

    **title** The text to be displayed on the button. Pass *None* if no title is desired.

    **callback** The method to be called when the user presses the button.

    **sizeStyle** A string representing the desired size style of the button.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+
    """

    nsButtonClass = NSButton
    nsBezelStyle = NSRoundedBezelStyle
    nsButtonType = NSMomentaryPushInButton

    frameAdjustments = {
        "mini": (-1, -2, 2, 2),
        "small": (-5, -7, 10, 11),
        "regular": (-6, -8, 12, 12),
        }

    def __init__(self, posSize, title, callback=None, sizeStyle="regular"):
        self._setupView(self.nsButtonClass, posSize, callback=callback)
        self._setSizeStyle(sizeStyle)
        self._nsObject.setTitle_(title)
        if self.nsBezelStyle is not None:
            self._nsObject.setBezelStyle_(self.nsBezelStyle)
        self._nsObject.setButtonType_(self.nsButtonType)

    def _testForDeprecatedAttributes(self):
        super(Button, self)._testForDeprecatedAttributes()
        from warnings import warn
        if hasattr(self, "_nsBezelStyle"):
            warn(DeprecationWarning("The _nsBezelStyle attribute is deprecated. Use the nsBezelStyle attribute."))
            self.nsBezelStyle = self._nsBezelStyle
        if hasattr(self, "_nsButtonType"):
            warn(DeprecationWarning("The _nsButtonType attribute is deprecated. Use the nsButtonType attribute."))
            self.nsButtonType = self._nsButtonType

    def getNSButton(self):
        """
        Return the `NSButton`_ that this object wraps.

        .. _NSButton: https://developer.apple.com/documentation/appkit/nsbutton?language=objc
        """
        return self._nsObject

    def bind(self, key, modifiers):
        """
        Bind a key to the button.

        **key** A single character or one of the following:

        +-----------------+
        | "help"          |
        +-----------------+
        | "home"          |
        +-----------------+
        | "end"           |
        +-----------------+
        | "pageup"        |
        +-----------------+
        | "pagedown"      |
        +-----------------+
        | "forwarddelete" |
        +-----------------+
        | "leftarrow"     |
        +-----------------+
        | "rightarrow"    |
        +-----------------+
        | "uparrow"       |
        +-----------------+
        | "downarrow"     |
        +-----------------+

        **modifiers** A list containing nothing or as many of the following as desired:

        +------------+
        | "command"  |
        +------------+
        | "control"  |
        +------------+
        | "option"   |
        +------------+
        | "shift"    |
        +------------+
        | "capslock" |
        +------------+
        """
        modifiers = sum([_modifierMap[i] for i in modifiers])
        key = _keyMap.get(key, key)
        self._nsObject.setKeyEquivalent_(key)
        self._nsObject.setKeyEquivalentModifierMask_(modifiers)


class SquareButton(Button):

    """
    A standard square button.

    .. image:: /_images/SquareButton.png

    ::

        from vanilla import *

        class SquareButtonDemo:

             def __init__(self):
                 self.w = Window((200, 100))
                 self.w.button = SquareButton((10, 10, -10, -10), "A Button",
                                    callback=self.buttonCallback)
                 self.w.open()

             def buttonCallback(self, sender):
                 print("button hit!")

        SquareButtonDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the button.

    **title** The text to be displayed on the button. Pass *None* if no title is desired.

    **callback** The method to be called when the user presses the button.

    **sizeStyle** A string representing the desired size style of the button.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+
    """

    nsBezelStyle = NSShadowlessSquareBezelStyle
    frameAdjustments = None

    def __init__(self, posSize, title, callback=None, sizeStyle="regular"):
        super(SquareButton, self).__init__(posSize=posSize, title=title, callback=callback, sizeStyle=sizeStyle)


_imagePositionMap = {
        "left": NSImageLeft,
        "right": NSImageRight,
        "top": NSImageAbove,
        "bottom": NSImageBelow,
        }

class ImageButton(SquareButton):

    """
    A button with an image.

    ::

        from vanilla import Window, ImageButton

        class ImageButtonDemo:

             def __init__(self):
                 path = "/path/to/an/image"
                 self.w = Window((50, 50))
                 self.w.button = ImageButton((10, 10, 30, 30), imagePath=path,
                                    callback=self.buttonCallback)
                 self.w.open()

             def buttonCallback(self, sender):
                 print("button hit!")

        ImageButtonDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the button.

    **title** The text to be displayed on the button. Pass *None* if no title is desired.

    **bordered** Boolean representing if the button should be bordered.

    **imagePath** A file path to an image.

    **imageNamed** The name of an image already loaded as a `NSImage`_ by the application.

    **imageObject** A `NSImage`_ object.

    .. note:: Only one of *imagePath*, *imageNamed*, *imageObject* should be set.

    **imagePosition** The position of the image relative to the title.
    The options are:

    +----------+
    | "top"    |
    +----------+
    | "bottom" |
    +----------+
    | "left"   |
    +----------+
    | "right"  |
    +----------+

    **callback** The method to be called when the user presses the button.

    **sizeStyle** A string representing the desired size style of the button.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+

    .. _NSImage: https://developer.apple.com/documentation/appkit/nsimage?language=objc
    """

    frameAdjustments = None

    def __init__(self, posSize,
                imagePath=None, imageNamed=None, imageObject=None,
                title=None, bordered=True, imagePosition="top", callback=None, sizeStyle="regular"):
        super(ImageButton,  self).__init__(posSize, title=title, callback=callback, sizeStyle=sizeStyle)
        image = None
        if imagePath is not None:
            image = NSImage.alloc().initWithContentsOfFile_(imagePath)
        elif imageNamed is not None:
            image = NSImage.imageNamed_(imageNamed)
        elif imageObject is not None:
            image = imageObject
        if image is not None:
            self._nsObject.setImage_(image)
        self._nsObject.setBordered_(bordered)
        if title is None:
            position = NSImageOnly
        else:
            position= _imagePositionMap[imagePosition]
            if imagePosition == "left":
                self._nsObject.setAlignment_(NSTextAlignmentRight)
            elif imagePosition == "right":
                self._nsObject.setAlignment_(NSTextAlignmentLeft)
        if not bordered:
            self._nsObject.cell().setHighlightsBy_(NSNoCellMask)
        self._nsObject.setImagePosition_(position)

    def setImage(self, imagePath=None, imageNamed=None, imageObject=None):
        """
        Set the image in the button.

        **imagePath** A file path to an image.

        **imageNamed** The name of an image already loaded as an `NSImage`_ by the application.

        **imageObject** A `NSImage`_ object.

        .. note:: Only one of *imagePath*, *imageNamed*, *imageObject* should be set.
        """
        if imagePath is not None:
            image = NSImage.alloc().initWithContentsOfFile_(imagePath)
        elif imageNamed is not None:
            image = NSImage.imageNamed_(imageNamed)
        elif imageObject is not None:
            image = imageObject
        else:
            raise ValueError("no image source defined")
        self._nsObject.setImage_(image)


class HelpButton(Button):

    """
    A standard help button.

    .. image:: /_images/HelpButton.png

    ::

        from vanilla import Window, HelpButton

        class HelpButtonDemo:

             def __init__(self):
                 self.w = Window((90, 40))
                 self.w.button = HelpButton((10, 10, 21, 20),
                                    callback=self.buttonCallback)
                 self.w.open()

             def buttonCallback(self, sender):
                 print("help button hit!")

        HelpButtonDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the button. The size of the button
    should match the standard dimensions.

    +-------------------------+
    | **Standard Dimensions** |
    +--------+----------------+
    | Width  | 21             |
    +--------+----------------+
    | Height | 20             |
    +--------+----------------+

    **callback** The method to be called when the user presses the button.
    """

    nsBezelStyle = NSHelpButtonBezelStyle
    frameAdjustments = {
        "regular": (0, -3, 0, 3),
        }

    def __init__(self, posSize, callback=None, page=None, anchor=None):
        # XXX perhaps this should choke if more than one arg is present.
        # callback, page and anchor are all mutually exclusive.
        self._page = page
        self._anchor = anchor
        if callback is None:
            if page is not None or anchor is not None:
                callback = self._helpBookCallback
        super(HelpButton, self).__init__(posSize, title="", callback=callback)

    def _helpBookCallback(self, sender):
        from Carbon import AH
        bundle = NSBundle.mainBundle()
        if bundle is None:
            return
        info = bundle.infoDictionary()
        helpBookName = info.get("CFBundleHelpBookName")
        if self._page is not None:
            AH.AHGoToPage(helpBookName, self._page, self._anchor)
        elif self._anchor is not None:
            AH.AHLookupAnchor(helpBookName, self._anchor)




from AppKit import NSTextField, NSSecureTextField
from vanilla.vanillaBase import VanillaBaseControl, VanillaCallbackWrapper


class VanillaEditTextDelegate(VanillaCallbackWrapper):

    _continuous = True

    def controlTextDidChange_(self, notification):
        if self._continuous:
            self.action_(notification.object())

    def controlTextDidEndEditing_(self, notification):
        if not self._continuous:
            self.action_(notification.object())


class EditText(VanillaBaseControl):

    """
    Standard short text entry control.

    .. image:: /_images/EditText.png

    ::

        from vanilla import Window, EditText

        class EditTextDemo:

            def __init__(self):
                self.w = Window((120, 42))
                self.w.editText = EditText((10, 10, -10, 22),
                                    text='some text',
                                    callback=self.editTextCallback)
                self.w.open()

            def editTextCallback(self, sender):
                print("text entry!", sender.get())

        EditTextDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the text entry control.

    +-------------------------+
    | **Standard Dimensions** |
    +---------+---+-----------+
    | Regular | H | 22        |
    +---------+---+-----------+
    | Small   | H | 19        |
    +---------+---+-----------+
    | Mini    | H | 16        |
    +---------+---+-----------+

    **text** An object representing the contents of the text entry control.
    If no formatter has been assigned to the control, this should be a string.
    If a formatter has been assigned, this should be an object of the type
    that the formatter expects.

    **callback** The method to be called when the user enters text.

    **continuous** If *True*, the callback (if any) will be called upon each
    keystroke, if *False*, only call the callback when editing finishes.
    Default is *True*.

    **readOnly** Boolean representing if the text can be edited or not.

    **formatter** A `NSFormatter`_ for controlling the display and input of
    the text entry.

    **placeholder** A placeholder string to be shown when the text entry
    control is empty.

    **sizeStyle** A string representing the desired size style of the text
    entry control. The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+

    .. _NSFormatter: https://developer.apple.com/documentation/foundation/nsformatter?language=objc
    """

    nsTextFieldClass = NSTextField
    nsTextFieldDelegateClass = VanillaEditTextDelegate

    def __init__(self, posSize, text="", callback=None, continuous=True, readOnly=False, formatter=None, placeholder=None, sizeStyle="regular"):
        self._continuous = continuous
        self._setupView(self.nsTextFieldClass, posSize, callback)
        self._posSize = posSize
        self._setSizeStyle(sizeStyle)
        self._nsObject.setObjectValue_(text)
        self._nsObject.setDrawsBackground_(True)
        self._nsObject.setBezeled_(True)
        self._nsObject.setEditable_(not readOnly)
        self._nsObject.setSelectable_(True)
        self._nsObject.setFormatter_(formatter)
        if placeholder:
            self._nsObject.cell().setPlaceholderString_(placeholder)

    def _testForDeprecatedAttributes(self):
        super(EditText, self)._testForDeprecatedAttributes()
        from warnings import warn
        if hasattr(self, "_textFieldClass"):
            warn(DeprecationWarning("The _textFieldClass attribute is deprecated. Use the nsTextFieldClass attribute."))
            self.nsTextFieldClass = self._textFieldClass

    def getNSTextField(self):
        """
        Return the `NSTextField`_ that this object wraps.

        .. _NSTextField: https://developer.apple.com/documentation/appkit/nstextfield?language=objc
        """
        return self._nsObject

    def _setCallback(self, callback):
        if callback is not None:
            self._target = self.nsTextFieldDelegateClass(callback)
            self._target._continuous = self._continuous
            self._nsObject.setDelegate_(self._target)

    def get(self):
        """
        Get the contents of the text entry control.

        If no formatter has been assigned to the control, this returns a string.
        If a formatter has been assigned, this returns an object which has been
        translated by the formatter.
        """
        return self._nsObject.objectValue()

    def set(self, value):
        """
        Set the contents of the text entry control.

        **value** An object representing the contents of the text entry control.
        If no formatter has been assigned to the control, this should be a string.
        If a formatter has been assigned, this should be an object of the type
        that the formatter expects.
        """
        self._nsObject.setObjectValue_(value)

    def getPlaceholder(self):
        """
        Get the placeholder string displayed in the control.
        """
        return self._nsObject.cell().placeholderString()

    def setPlaceholder(self, value):
        """
        Set *value* as the placeholder string to be displayed in the control.
        *value* must be a string.
        """
        self._nsObject.cell().setPlaceholderString_(value)

    def selectAll(self):
        """
        Select all text in the text entry control.
        """
        self._nsObject.selectText_(None)


class SecureEditText(EditText):

    """
    Standard secure text entry control.

    .. image:: /_images/SecureEditText.png

    ::

        from vanilla import Window, SecureEditText

        class SecureEditTextDemo:

            def __init__(self):
                self.w = Window((120, 42))
                self.w.secureEditText = SecureEditText((10, 10, -10, 22),
                                    text='abc123',
                                    callback=self.secureEditTextCallback)
                self.w.open()

            def secureEditTextCallback(self, sender):
                print("text entry!", sender.get())

        SecureEditTextDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the text entry control.

    +-------------------------+
    | **Standard Dimensions** |
    +---------+---+-----------+
    | Regular | H | 22        |
    +---------+---+-----------+
    | Small   | H | 19        |
    +---------+---+-----------+
    | Mini    | H | 16        |
    +---------+---+-----------+

    **text** An object representing the contents of the text entry control.
    If no formatter has been assigned to the control, this should be a string.
    If a formatter has been assigned, this should be an object of the type
    that the formatter expects.

    **callback** The method to be called when the user enters text.

    **continuous** If *True*, the callback (if any) will be called upon each
    keystroke, if *False*, only call the callback when editing finishes.
    Default is *True*.

    **readOnly** Boolean representing if the text can be edited or not.

    **formatter** A `NSFormatter`_ for controlling the display and input
    of the text entry.

    **placeholder** A placeholder string to be shown when the text  entry
    control is empty.

    **sizeStyle** A string representing the desired size style of the text
    entry control. The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+

    .. _NSFormatter: https://developer.apple.com/documentation/foundation/nsformatter?language=objc
    """

    nsTextFieldClass = NSSecureTextField

    def getNSSecureTextField(self):
        """
        Return the `NSSecureTextField`_ that this object wraps.

        .. _NSSecureTextField: https://developer.apple.com/documentation/appkit/nssecuretextfield?language=objc
        """
        return self._nsObject



#!/usr/bin/env python

# import sys
from setuptools import setup


# if "sdist" in sys.argv:
#     import os
#     import subprocess
#     import shutil
#     docFolder = os.path.join(os.getcwd(), "Documentation")
#     # remove existing
#     doctrees = os.path.join(docFolder, "build", "doctrees")
#     if os.path.exists(doctrees):
#         shutil.rmtree(doctrees)
#     # compile
#     p = subprocess.Popen(["make", "html"], cwd=docFolder)
#     p.wait()
#     # remove doctrees
#     shutil.rmtree(doctrees)


with open("README.md", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="cocoa-vanilla",
    use_scm_version={"write_to": "Lib/vanilla/_version.py"},
    description="A Pythonic wrapper around Cocoa.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Tal Leming",
    author_email="tal@typesupply.com",
    maintainer="Just van Rossum, Frederik Berlaen",
    maintainer_email="justvanrossum@gmail.com",
    install_requires=["pyobjc"],
    setup_requires=["setuptools_scm"],
    url="https://github.com/robotools/vanilla",
    license="MIT",
    packages=["vanilla", "vanilla.test"],
    package_dir={"": "Lib"},
    platforms=["macOS"],
    python_requires=">=3.6",
)



from Foundation import NSDate
from AppKit import NSRunLoop, NSProgressIndicator, NSProgressIndicatorBarStyle, NSProgressIndicatorSpinningStyle
from vanilla.vanillaBase import VanillaBaseObject, _sizeStyleMap, osVersion10_11, osVersionCurrent

class ProgressBar(VanillaBaseObject):

    """
    A standard progress bar.

    .. image:: /_images/ProgressBar.png

    ::

        from vanilla import Window, ProgressBar, Button

        class ProgressBarDemo:

            def __init__(self):
                self.w = Window((200, 65))
                self.w.bar = ProgressBar((10, 10, -10, 16))
                self.w.button = Button((10, 35, -10, 20), "Go!",
                                    callback=self.showProgress)
                self.w.open()

            def showProgress(self, sender):
                import time
                self.w.bar.set(0)
                for i in range(10):
                    self.w.bar.increment(10)
                    time.sleep(.2)

        ProgressBarDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the progress bar. The height of the
    progress bar should match the appropriate value for the given *sizeStyle*.

    +---------------------------+
    | **Standard Dimensions**   |
    +-----------+---+-----------+
    | *Regular* | H | 16        |
    +-----------+---+-----------+
    | *Small*   | H | 10        |
    +-----------+---+-----------+

    **minValue** The minimum value of the progress bar.

    **maxValue** The maximum value of the progress bar.

    **isIndeterminate** Boolean representing if the progress bar is
    indeterminate. Determinate progress bars show how much of the task
    has been completed. Indeterminate progress bars simply show that
    the application is busy.

    **sizeStyle** A string representing the desired size style of the
    pregress bar. The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    """

    nsProgressIndicatorClass = NSProgressIndicator

    allFrameAdjustments = {
        "small": (-1, -0, 2, 0),
        "regular": (-2, -0, 4, 0),
    }

    progressStyleMap = {
        "bar" : NSProgressIndicatorBarStyle,
        "spinning" : NSProgressIndicatorSpinningStyle
    }

    def __init__(self, posSize, minValue=0, maxValue=100, isIndeterminate=False, sizeStyle="regular", progressStyle="bar"):
        self._setupView(self.nsProgressIndicatorClass, posSize)
        self.frameAdjustments = self.allFrameAdjustments[sizeStyle]
        self._nsObject.setControlSize_(_sizeStyleMap[sizeStyle])
        self._nsObject.setStyle_(self.progressStyleMap[progressStyle])
        self._nsObject.setMinValue_(minValue)
        self._nsObject.setMaxValue_(maxValue)
        self._nsObject.setIndeterminate_(isIndeterminate)
        if isIndeterminate:
            self._nsObject.setUsesThreadedAnimation_(True)

    def getNSProgressIndicator(self):
        """
        Return the `NSProgressIndicator`_ that this object wraps.

        .. _NSProgressIndicator: https://developer.apple.com/documentation/appkit/nsprogressindicator?language=objc
        """
        return self._nsObject

    # implementation note:
    # display is called manually to ensure that the animation happens.
    # otherwise if it is called during an expensive python function,
    # it will not animate until the function is complete.

    def set(self, value):
        """
        Set the value of the progress bar to **value**.

        *Only available in determinate progress bars.*
        """
        self._nsObject.setDoubleValue_(value)
        self._nsObject.display()
        if osVersionCurrent >= osVersion10_11:
            NSRunLoop.mainRunLoop().runUntilDate_(NSDate.dateWithTimeIntervalSinceNow_(0.0001))

    def get(self):
        """
        Get the current value of the progress bar.

        *Only available in determinate progress bars.*
        """
        return self._nsObject.doubleValue()

    def increment(self, value=1):
        """
        Increment the progress bar by **value**.

        *Only available in determinate progress bars.*
        """
        self._nsObject.incrementBy_(value)
        self._nsObject.display()
        if osVersionCurrent >= osVersion10_11:
            NSRunLoop.mainRunLoop().runUntilDate_(NSDate.dateWithTimeIntervalSinceNow_(0.0001))

    def start(self):
        """
        Start the animation.

        *Only available in indeterminate progress bars.*
        """
        self._nsObject.startAnimation_(None)

    def stop(self):
        """
        Stop the animation.

        *Only available in indeterminate progress bars.*
        """
        self._nsObject.stopAnimation_(None)



from AppKit import NSSlider, NSTickMarkLeft, NSTickMarkRight, NSTickMarkAbove, NSTickMarkBelow
from vanilla.vanillaBase import VanillaBaseControl, VanillaError, osVersionCurrent, osVersion10_12

_tickPositionMap = {
    "left": NSTickMarkLeft,
    "right": NSTickMarkRight,
    "top": NSTickMarkAbove,
    "bottom": NSTickMarkBelow,
}


class Slider(VanillaBaseControl):

    """
    A standard slider control. Sliders can be vertical or horizontal
    and they can show tick marks or not.

    .. image:: /_images/Slider.png

    ::

        from vanilla import Window, Slider

        class SliderDemo:

             def __init__(self):
                 self.w = Window((200, 43))
                 self.w.slider = Slider((10, 10, -10, 23),
                                    tickMarkCount=10,
                                    callback=self.sliderCallback)
                 self.w.open()

             def sliderCallback(self, sender):
                 print("slider edit!", sender.get())

        SliderDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the slider. The size of the slider
    should match the appropriate value for the given *sizeStyle*.

    +---------------------------+
    | **Standard Dimensions**   |
    +---------------------------+
    | *without ticks*           |
    +---------+---+----+---+----+
    | Regular | W | 15 | H | 15 |
    +---------+---+----+---+----+
    | Small   | W | 12 | H | 11 |
    +---------+---+----+---+----+
    | Mini    | W | 10 | H | 10 |
    +---------+---+----+---+----+
    | *with ticks*              |
    +---------+---+----+---+----+
    | Regular | W | 24 | H | 23 |
    +---------+---+----+---+----+
    | Small   | W | 17 | H | 17 |
    +---------+---+----+---+----+
    | Mini    | W | 16 | H | 16 |
    +---------+---+----+---+----+

    **minValue** The minimum value allowed by the slider.

    **maxValue** The maximum value allowed by the slider.

    **value** The initial value of the slider.

    **tickMarkCount** The number of tick marcks to be displayed on the slider.
    If *None* is given, no tick marks will be displayed.

    **stopOnTickMarks** Boolean representing if the slider knob should only
    stop on the tick marks.

    **continuous** Boolean representing if the assigned callback should be
    called during slider editing. If *False* is given, the callback will be
    called after the editing has finished.

    **callback** The method to be called when the slider has been edited.

    **sizeStyle** A string representing the desired size style of the slider.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+
    """

    nsSliderClass = NSSlider

    allFrameAdjustments = {
        "H-Slider-Above": {
            "mini": (0, 0, 0, 0),
            "small": (0, -1, 0, -1),
            "regular": (-2, -2, 4, 2),
        },
        "H-Slider-Below": {
            "mini": (0, 0, 0, 0),
            "small": (0, 0, 0, 0),
            "regular": (-2, 0, 4, 1),
        },
        "H-Slider-None": {
            "mini": (0, -1, 0, 2),
            "small": (0, -2, 0, 3),
            "regular": (-2, -4, 4, 6),
        },
        "V-Slider-Left": {
            "mini": (0, -1, 1, 1),
            "small": (0, -1, 1, 1),
            "regular": (0, -3, 2, 5),
        },
        "V-Slider-Right": {
            "mini": (0, -1, 1, 1),
            "small": (-1, -1, 2, 1),
            "regular": (-2, -3, 2, 5),
        },
        "V-Slider-None": {
            "mini": (0, -1, 1, 1),
            "small": (-2, -1, 4, 1),
            "regular": (-3, -3, 6, 5),
        },
    }

    def __init__(self, posSize, minValue=0, maxValue=100, value=50,
            tickMarkCount=None, stopOnTickMarks=False, continuous=True,
            callback=None, sizeStyle="regular"):
        self._setupView(self.nsSliderClass, posSize, callback=callback)
        self._setSizeStyle(sizeStyle)
        self._nsObject.setMinValue_(minValue)
        self._nsObject.setMaxValue_(maxValue)
        self._nsObject.setFloatValue_(value)
        if tickMarkCount:
            self._nsObject.setNumberOfTickMarks_(tickMarkCount)
            if stopOnTickMarks:
                self._nsObject.setAllowsTickMarkValuesOnly_(True)
        if continuous:
            self._nsObject.setContinuous_(True)
        else:
            self._nsObject.setContinuous_(False)

    def getNSSlider(self):
        """
        Return the `NSSlider`_ that this object wraps.

        .. _NSSlider: https://developer.apple.com/documentation/appkit/nsslider?language=objc
        """
        return self._nsObject

    def _isVerticalFromFrame(self, frame):
        w, h = frame[1]
        return w < h

    def _adjustPosSize(self, frame):
        isVertical = self._isVerticalFromFrame(frame)
        if osVersionCurrent >= osVersion10_12:
            self._nsObject.setVertical_(isVertical)
        if isVertical:
            prefix = "V-"
        else:
            prefix = "H-"
        tickPos = "None"
        tickMarkCount = self._nsObject.numberOfTickMarks()
        if tickMarkCount:
            tickPos = self._nsObject.tickMarkPosition()
            if isVertical:
                if tickPos == NSTickMarkLeft:
                    tickPos = "Left"
                elif tickPos == NSTickMarkRight:
                    tickPos = "Right"
                # during __init__, the _nsObject will be unable
                # to determine if the slider is horizontal or
                # vertical, so it will return the position for
                # horizontal sliders. override that and default
                # to right here.
                else:
                    tickPos = "Right"
            else:
                if tickPos == NSTickMarkBelow:
                    tickPos = "Below"
                elif tickPos == NSTickMarkAbove:
                    tickPos = "Above"
        sliderType = prefix + "Slider-" + tickPos
        self.frameAdjustments = self.allFrameAdjustments[sliderType]
        # now let the super class do the work
        return super(Slider, self)._adjustPosSize(frame)

    def get(self):
        """
        Get the value of the slider.
        """
        return self._nsObject.floatValue()

    def set(self, value):
        """
        Set the value of the slider.
        """
        self._nsObject.setFloatValue_(value)

    def setMinValue(self, value):
        """
        Set the minimum value allowed by the slider.
        """
        self._nsObject.setMinValue_(value)

    def setMaxValue(self, value):
        """
        Set the maximum value allowed by the slider.
        """
        self._nsObject.setMaxValue_(value)

    def setTickMarkCount(self, value):
        """
        Set the number of tick marks on the slider.
        """
        self._nsObject.setNumberOfTickMarks_(value)

    def setTickMarkPosition(self, value):
        """
        Set the position of the tick marks on the slider.

        For vertical sliders, the options are:

        +---------+
        | "left"  |
        +---------+
        | "right" |
        +---------+

        For horizontal sliders, the options are:

        +----------+
        | "top"    |
        +----------+
        | "bottom" |
        +----------+
        """
        position = _tickPositionMap[value]
        self._nsObject.setTickMarkPosition_(position)



from vanilla.vanillaBase import VanillaBaseObject
import AppKit

NSUserInterfaceLayoutOrientationHorizontal = 0
NSUserInterfaceLayoutOrientationVertical = 1

NSStackViewGravityTop = 1
NSStackViewGravityLeading = 1
NSStackViewGravityCenter = 2
NSStackViewGravityBottom = 3
NSStackViewGravityTrailing = 3

distributions = dict(
    equalCentering=AppKit.NSStackViewDistributionEqualCentering,
    equalSpacing=AppKit.NSStackViewDistributionEqualSpacing,
    fill=AppKit.NSStackViewDistributionFill,
    fillEqually=AppKit.NSStackViewDistributionFillEqually,
    fillProportionally=AppKit.NSStackViewDistributionFillProportionally,
    gravity=AppKit.NSStackViewDistributionGravityAreas,
    # deprecated
    gravityAreas=AppKit.NSStackViewDistributionGravityAreas
)


_doc = """
A view that allows the creation of a stack of views.

**posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
the position and size of the stack view.

**views** The views to display in the stack. See below for structure.

**spacing** Space to insert between views.

**alignment** The alignment of the views. Options:

* "leading"
* "center"
* "trailing"
* One of the `NSLayoutAttribute`_ options.

**distribution** The distribution of the views. Options:

* "equalCentering"
* "equalSpacing"
* "fill"
* "fillEqually"
* "fillProportionally"
* "gravity"

**edgeInsets** Tuple of four numbers (left, top, right, bottom) indicating the amount to inset the views.

**View Definition Structure**

Views are defined with either a Vanilla object, a NSView
(or NSView subclass) object or a dictionary with this structure:

* **view** A vanilla object or an instance of `NSView`_.
* **width** A  number, string (using the size syntax below)
  or None (indicating auto width) defining the width.
* **height** A  number, string (using the size syntax below)
  or None (indicating auto height) defining the height.
* **gravity** The gravity that this view should be attracted to.
* **spacing** A number defining custom spacing after the view.

Size Constants:

* "fill" Stretch the view to fit the width or height of
  the stack view.

Size Syntax:

* "==value" where value can be coerced to an integer or float
* "<=value" where value can be coerced to an integer or float
* ">=value" where value can be coerced to an integer or float

Up to two are allowed. Separate with ``,`` (comma).

Horizontal Gravity Options:

* "leading"
* "center"
* "trailing"
* One of the `NSStackViewGravity`_ options.

Vertical Gravity Options:

* "top"
* "center"
* "bottom"
* One of the `NSStackViewGravity`_ options.

.. _NSLayoutAttribute: https://developer.apple.com/documentation/uikit/nslayoutattribute?language=objc
.. _NSView: https://developer.apple.com/documentation/appkit/nsview?language=objc
.. _NSStackViewGravity: https://developer.apple.com/documentation/appkit/nsstackviewgravity?language=occ
"""

class _StackView(VanillaBaseObject):

    __doc__ = _doc

    nsStackViewClass = AppKit.NSStackView
    _orientation = None
    _gravities = None
    _alignments = None

    def __init__(self, posSize, views, spacing=0, alignment="center", distribution="fillEqually", edgeInsets=(0, 0, 0, 0)):
        self._setupView(self.nsStackViewClass, posSize)
        alignment = self._alignments.get(alignment, alignment)
        distribution = distributions[distribution]
        stackView = self._getContentView()
        stackView.setOrientation_(self._orientation)
        stackView.setSpacing_(spacing)
        stackView.setAlignment_(alignment)
        stackView.setDistribution_(distribution)
        for view in views:
            if not isinstance(view, dict):
                view = dict(view=view)
            self.appendView(**view)
        self.setEdgeInsets(edgeInsets)

    def getNSStackView(self):
        return self._nsObject

    def setEdgeInsets(self, value):
        """
        Set the edge insets.
        """
        stackView = self.getNSStackView()
        stackView.setEdgeInsets_(value)

    def appendView(self, view, width=None, height=None, spacing=None, gravity="center"):
        """
        Append a view.
        """
        index = len(self._nsObject.viewsInGravity_(self._gravities.get(gravity, gravity)))
        self.insertView(index, view=view, width=width, height=height, spacing=spacing, gravity=gravity)

    def insertView(self, index, view, width=None, height=None, spacing=None, gravity="center"):
        """
        Insert a view.
        """
        gravity = self._gravities.get(gravity, gravity)
        if isinstance(view, VanillaBaseObject):
            view = view._nsObject
        stackView = self.getNSStackView()
        stackView.insertView_atIndex_inGravity_(view, index, gravity)
        if isinstance(view, VanillaBaseObject):
            view = view._nsObject
        stackView = self.getNSStackView()
        widthHuggingPriority = AppKit.NSLayoutPriorityRequired
        widthCompressionPriority = AppKit.NSLayoutPriorityRequired
        if width is None:
            intrinsicWidth = view.intrinsicContentSize()[0]
            if intrinsicWidth == AppKit.NSViewNoInstrinsicMetric:
                widthHuggingPriority = AppKit.NSLayoutPriorityDefaultLow
                widthCompressionPriority = AppKit.NSLayoutPriorityDefaultLow
        elif isinstance(width, (float, int)):
            pass
        elif isinstance(width, str):
            if width == "fill":
                widthHuggingPriority = AppKit.NSLayoutPriorityDefaultLow
                widthCompressionPriority = AppKit.NSLayoutPriorityDefaultLow
            elif width.startswith("=="):
                pass
            elif width.startswith(">="):
                widthHuggingPriority = AppKit.NSLayoutPriorityDefaultLow
            elif width.startswith("<="):
                widthCompressionPriority = AppKit.NSLayoutPriorityDefaultLow
        view.setContentHuggingPriority_forOrientation_(
            widthHuggingPriority,
            AppKit.NSLayoutConstraintOrientationHorizontal
        )
        view.setContentCompressionResistancePriority_forOrientation_(
            widthCompressionPriority,
            AppKit.NSLayoutConstraintOrientationHorizontal
        )
        heightHuggingPriority = AppKit.NSLayoutPriorityRequired
        heightCompressionPriority = AppKit.NSLayoutPriorityRequired
        if height is None:
            intrinsicHeight = view.intrinsicContentSize()[1]
            if intrinsicHeight == AppKit.NSViewNoInstrinsicMetric:
                heightHuggingPriority = AppKit.NSLayoutPriorityDefaultLow
                heightCompressionPriority = AppKit.NSLayoutPriorityDefaultLow
        elif isinstance(height, (float, int)):
            pass
        elif isinstance(height, str):
            if height == "fill":
                heightHuggingPriority = AppKit.NSLayoutPriorityDefaultLow
                heightCompressionPriority = AppKit.NSLayoutPriorityDefaultLow
            elif height.startswith("=="):
                pass
            elif height.startswith(">="):
                heightHuggingPriority = AppKit.NSLayoutPriorityDefaultLow
            elif height.startswith("<="):
                heightCompressionPriority = AppKit.NSLayoutPriorityDefaultLow
        view.setContentHuggingPriority_forOrientation_(
            heightHuggingPriority,
            AppKit.NSLayoutConstraintOrientationVertical
        )
        view.setContentCompressionResistancePriority_forOrientation_(
            heightCompressionPriority,
            AppKit.NSLayoutConstraintOrientationVertical
        )
        if width is not None:
            if width == "fill":
                pass
            else:
                _applyStackViewConstantToAnchor(view.widthAnchor(), width)
        if height is not None:
            if height == "fill":
                pass
            else:
                _applyStackViewConstantToAnchor(view.heightAnchor(), height)
        if spacing is not None:
            stackView.setCustomSpacing_afterView_(spacing, view)

    def removeView(self, view):
        """
        Remove a view.
        """
        if isinstance(view, VanillaBaseObject):
            view = view._nsObject
        self.getNSStackView().removeView_(view)


class HorizontalStackView(_StackView):

    __doc__ = _doc

    _orientation = NSUserInterfaceLayoutOrientationHorizontal
    _gravities = dict(
        leading=NSStackViewGravityLeading,
        center=NSStackViewGravityCenter,
        trailing=NSStackViewGravityTrailing
    )
    _alignments = dict(
        center=AppKit.NSLayoutAttributeCenterY,
        leading=AppKit.NSLayoutAttributeTop,
        trailing=AppKit.NSLayoutAttributeBottom,
        height=AppKit.NSLayoutAttributeHeight
    )

class VerticalStackView(_StackView):

    __doc__ = _doc

    _orientation = NSUserInterfaceLayoutOrientationVertical
    _gravities = dict(
        top=NSStackViewGravityTop,
        center=NSStackViewGravityCenter,
        bottom=NSStackViewGravityBottom
    )
    _alignments = dict(
        center=AppKit.NSLayoutAttributeCenterX,
        leading=AppKit.NSLayoutAttributeLeading,
        trailing=AppKit.NSLayoutAttributeTrailing,
        width=AppKit.NSLayoutAttributeWidth
    )


def _applyStackViewConstantToAnchor(anchor, value):
    for existing in anchor.constraintsAffectingLayout():
        existing.setActive_(False)
    if isinstance(value, str) and "," in value:
        for v in value.split(","):
            v = v.strip()
            _applyStackViewConstantToAnchor(anchor, v)
        return
    methods = {
        "==" : anchor.constraintEqualToConstant_,
        ">=" : anchor.constraintGreaterThanOrEqualToConstant_,
        "<=" : anchor.constraintLessThanOrEqualToConstant_,
    }
    if isinstance(value, (int, float)):
        method = methods["=="]
    else:
        relation = value[:2]
        value = float(value[2:])
        method = methods[relation]
    constraint = method(value)
    constraint.setActive_(True)



from AppKit import NSPopUpButton, NSPopUpButtonCell, NSMenuItem, NSImageNameActionTemplate, NSImage, NSMenu, NSTexturedRoundedBezelStyle
from vanilla.vanillaBase import VanillaBaseControl, VanillaCallbackWrapper, _reverseSizeStyleMap
from vanilla.vanillaList import VanillaMenuBuilder


class PopUpButton(VanillaBaseControl):

    """
    A button which, when selected, displays a list of items for the user to choose from.

    .. image:: /_images/PopUpButton.png

    ::

        from vanilla import Window, PopUpButton

        class PopUpButtonDemo:

            def __init__(self):
                self.w = Window((100, 40))
                self.w.popUpButton = PopUpButton((10, 10, -10, 20),
                                      ["A", "B", "C"],
                                      callback=self.popUpButtonCallback)
                self.w.open()

            def popUpButtonCallback(self, sender):
                print("pop up button selection!", sender.get())

        PopUpButtonDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the pop up button. The size of the button should match
    the appropriate value for the given *sizeStyle*.

    +---------------------------+
    | **Standard Dimensions**   |
    +-----------+---+-----------+
    | *Regular* | H | 20        |
    +-----------+---+-----------+
    | *Small*   | H | 17        |
    +-----------+---+-----------+
    | *Mini*    | H | 15        |
    +-----------+---+-----------+

    **items** A list of items to appear in the pop up list.

    **callback** The method to be called when the user selects an item in the
    pop up list.

    **sizeStyle** A string representing the desired size style of the pop up button.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+
    """

    nsPopUpButtonClass = NSPopUpButton
    nsPopUpButtonCellClass = NSPopUpButtonCell

    frameAdjustments = {
        "mini": (-1, 0, 3, 0),
        "small": (-3, -4, 6, 5),
        "regular": (-3, -4, 6, 6),
    }

    def __init__(self, posSize, items, callback=None, sizeStyle="regular"):
        self._setupView(self.nsPopUpButtonClass, posSize)
        if self.nsPopUpButtonCellClass != NSPopUpButtonCell:
            self._nsObject.setCell_(self.nsPopUpButtonCellClass.alloc().init())
        if callback is not None:
            self._setCallback(callback)
        self._setSizeStyle(sizeStyle)
        self.setItems(items)

    def getNSPopUpButton(self):
        """
        Return the `NSPopUpButton`_ that this object wraps.

        .. _NSPopUpButton: https://developer.apple.com/documentation/appkit/nspopupbutton?language=objc
        """
        return self._nsObject

    def get(self):
        """
        Get the index selected item in the pop up list.
        """
        return self._nsObject.indexOfSelectedItem()

    def set(self, value):
        """
        Set the index of the selected item in the pop up list.
        """
        self._nsObject.selectItemAtIndex_(value)

    def getItem(self):
        """
        Get the selected item title in the pop up list.
        """
        return self._nsObject.titleOfSelectedItem()

    def setItem(self, item):
        """
        Set the selected item title in the pop up list.
        """
        self._nsObject.selectItemWithTitle_(item)

    def setItems(self, items):
        """
        Set the items to appear in the pop up list.
        """
        self._nsObject.removeAllItems()
        for item in items:
            if isinstance(item, NSMenuItem):
                menu = self._nsObject.menu()
                menu.addItem_(item)
            else:
                self._nsObject.addItemWithTitle_(item)

    def getItems(self):
        """
        Get the list of items that appear in the pop up list.
        """
        return self._nsObject.itemTitles()


class ActionButton(PopUpButton):

    """
    An action button with a menu.

    .. image:: /_images/ActionButton.png

    ::

        from vanilla import Window, ActionButton

        class ActionPopUpButtonDemo:

            def __init__(self):
                self.w = Window((100, 40))
                items = [
                        dict(title="first", callback=self.firstCallback),
                        dict(title="second", callback=self.secondCallback),
                        dict(title="third", items=[
                                dict(title="sub first", callback=self.subFirstCallback)
                            ])
                        ]
                self.w.actionPopUpButton = ActionButton((10, 10, 30, 20), items)
                self.w.open()

            def firstCallback(self, sender):
                print("first")

            def secondCallback(self, sender):
                print("second")

            def subFirstCallback(self, sender):
                print("sub first")

        ActionPopUpButtonDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the pop up button. The size of the button should match
    the appropriate value for the given *sizeStyle*.

    +---------------------------+
    | **Standard Dimensions**   |
    +-----------+---+-----------+
    | *Regular* | H | 20        |
    +-----------+---+-----------+
    | *Small*   | H | 17        |
    +-----------+---+-----------+
    | *Mini*    | H | 15        |
    +-----------+---+-----------+

    **items** A list of items to appear in the pop up list as dictionaries.
    Optionally an item could be a `NSMenuItem`_.
    When an item is set to ``----``, it will be a menu item separator.

    +------------+----------------------------------------------+
    | "title"    | The title of the item.                       |
    +------------+----------------------------------------------+
    | "callback" | The callback of the item.                    |
    +------------+----------------------------------------------+
    | "items"    | Each item could have sub menus, as list of   |
    |            | dictionaries with the same format.           |
    +------------+----------------------------------------------+

    **sizeStyle** A string representing the desired size style of the pop up button.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+

    **bordered** Boolean representing if the button should be bordered.

    .. _NSMenuItem: https://developer.apple.com/documentation/appkit/nsmenuitem?language=objc
    """

    def __init__(self, posSize, items, sizeStyle="regular", bordered=True):
        super(ActionButton, self).__init__(posSize, items, sizeStyle=sizeStyle)
        self._nsObject.setPullsDown_(True)
        self._nsObject.setBezelStyle_(NSTexturedRoundedBezelStyle)
        self._nsObject.setBordered_(bordered)

    def _breakCycles(self):
        self._menuItemCallbackWrappers = None
        super(ActionButton, self)._breakCycles()

    def getFirstItem(self):
        actionImage = NSImage.imageNamed_(NSImageNameActionTemplate).copy()
        sizeStyle = _reverseSizeStyleMap[self._nsObject.cell().controlSize()]
        if sizeStyle == "small":
            actionImage.setSize_((10, 10))
        elif sizeStyle == "mini":
            actionImage.setSize_((10, 10))
        firstActionItem = NSMenuItem.alloc().init()
        firstActionItem.setImage_(actionImage)
        firstActionItem.setTitle_("")
        return firstActionItem

    def setItems(self, items):
        """
        Set the items to appear in the pop up list.
        """
        self._callbackWrappers = []
        self._nsObject.removeAllItems()
        menu = NSMenu.alloc().init()
        menu.addItem_(self.getFirstItem())
        VanillaMenuBuilder(self, items, menu)
        self._nsObject.setMenu_(menu)
